# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from entangled.generic.analyser import Graph, Block
from entangled.op_types import ProjType, OpType
from entangled.operation.definitions import SSAElement, Proj


class GenericSelfish(object):
    def __init__(self, constructor, data, base=None):
        self.base = base
        self.constructor = constructor
        self.data = data


class SilentEventHandler(object):
    def handle_ob(self, result, target, event_name, *args):
        pass

    def handle(self, target, event_name, *args):
        pass
            

class JsonEventHandler(object):
    def __init__(self, transport):
        self._value_table = {}
        self.transport = transport

    def _encode_selfish(self, identity, name, data, base=None):
        try:
            return self._value_table[identity]
        except KeyError:
            self.handle_ob(identity, base, name, *data)
            return self._value_table[identity]

    def _encode_optype(self, ty):
        return self._encode_selfish(ty, 'optype', [
            ty.name, ty.args, ty.projs, ty.opcls, ty.stararg, ty.is_phi
        ])

    def _encode_projtype(self, ty):
        return self._encode_selfish(ty, 'projtype', [ty.name, ty.mode])

    def _encode_proj(self, proj):
        return {
            'parent': self._value_table[proj.parent],
            'type': self._encode_projtype(proj.proj),
            'index': proj.index,
        }

    def _encode(self, item):
        if isinstance(item, Proj):
            return self._encode_proj(item)
        if isinstance(item, ProjType):
            return self._encode_projtype(item)
        if isinstance(item, OpType):
            return self._encode_optype(item)
        if isinstance(item, GenericSelfish):
            return self._encode_selfish(item, item.constructor, item.data,
                                        item.base)
        if isinstance(item, (Graph, Block, SSAElement)):
            return self._value_table[item]
        if isinstance(item, (str, unicode, int, float, type(None))):
            return item
        if isinstance(item, dict):
            return dict((k, self._encode(v)) for k, v in item.iteritems())
        return self._encode_list(item)
            
    def _encode_list(self, items):
        return [self._encode(item) for item in items]

    def send(self, target, verb, args=(), save=None):
        if target is not None:
            target = self._value_table[target]
        encoded_args = self._encode_list(args)

        question_id = None
        if save is not None:
            question_id = len(self._value_table)
            self._value_table[save] = question_id
        self.transport.send(json.dumps([question_id, target, verb,
                                        encoded_args]))

    def handle_ob(self, result, target, event_name, *args):
        self.send(target, event_name, args, result)

    def handle(self, target, event_name, *args):
        self.send(target, event_name, args)
            
