# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class Layout(object):
    def __init__(self):
        self.column_use = {}
        self.node_data = {}
        self.data_tabular = []
        self.value_usage = {}
        self.all_phi_nodes = []

    def get_cell(self, node):
        data = self.node_data[node]
        return data['row'], data['column']

    def layout(self, node):
        if node.is_phi:
            self.layout_phi(node)
        elif node.is_op:
            if node.is_proj:
                self.layout_proj(node)
            else:
                self.layout_node(node)
        elif node.is_list:
            self.layout_node(node)

    def store(self, node, row, column, constants):
        data = {'row' : row, 'column' : column, 'constants' : constants,
                'proj' : []}
        self.node_data[node] = data
        if row >= len(self.data_tabular):
            assert row == len(self.data_tabular)
            self.data_tabular.append({})
        self.data_tabular[row][column] = node, data

    def layout_phi(self, node):
        rendered_args = [arg for arg in node.all_args
                         if arg in self.node_data]
        if not rendered_args:
            assert all(arg.is_constant for arg in node.all_args
            ), "%r\n%r" % (node, node.all_args)
        self.all_phi_nodes.append(node)
        self.layout_base(node, rendered_args)

    def layout_node(self, node):
        op_args = [arg for arg in node.all_args if arg in self.node_data]
        for arg in node.all_args:
            if not arg.is_constant:
                self.value_usage.setdefault(arg, []).append(node)
#                if not arg.is_phi:
#                    op_args.append(arg)
        self.layout_base(node, op_args)

    def layout_proj(self, node):
        target = node.parent
        target_data = self.node_data[target]
        try:
            target_data['proj'].append(node)
        except KeyError:
            print target_data
            raise
        self.node_data[node] = {'target' : target,
                                'row' : target_data['row'],
                                'column' : target_data['column'],
                                'proj' : []}
        
    def layout_base(self, node, args):
        row = column = 0
        if args:
            row = 1 + max(self.node_data[arg]['row'] for arg in args)
            column = min(self.node_data[arg]['column'] for arg in args)
        avail_columns = self.column_use.setdefault(row, set())
        while column in avail_columns:
            column += 1
        avail_columns.add(column)
        constants = [arg for arg in node.arguments if arg.is_constant]
        self.store(node, row, column, constants)


def allocate_crosslinks(liveness):
    allocation = []
    column_allocation = {}

    for row, variables in sorted(liveness.items()):
        if set(allocation) == set(variables):
            continue
        for index, v in enumerate(allocation):
            if v not in variables:
                allocation[index] = None
        for v in variables:
            if v in allocation:
                continue
            if None in allocation:
                index = allocation.index(None)
                allocation[index] = v
                column_allocation[v] = index
            else:
                column_allocation[v] = len(allocation)
                allocation.append(v)
    return len(allocation), column_allocation


def compute_liveness(layout):
    """build a map of variable liveness for each cell.

    Each variable is allocated a column, and within this column, it is
    only live for certain rows.  The first map is a structure that
    reflects this: For each column, it is a map from the row to the
    set of variables live at that row.

    Additionally, within each row, there are variables which need to
    travel from their source to each destination.  The second map is,
    for each row, a map from the column to the number of variables
    live in that column.

    This function returns these two maps: one describing liveness
    within each column, and one describing liveness across each row.

    """
    liveness = {}   # {column : {row : {variable}}}
    liveness_t = {} # {row : {column : {variable}}}
    for source, targets in layout.value_usage.iteritems():
        start_row, col_source = layout.get_cell(source)
        max_row = max(layout.node_data[t]['row'] for t in targets)
        column_liveness = liveness.setdefault(col_source, {})
        for row in xrange(start_row, max_row + 1):
            column_liveness.setdefault(row, set()).add(source)
        for target in targets:
            target_data = layout.node_data[target]
            assert target_data['row'] > start_row, "expected %s > %s" % (
                target_data['row'], start_row)
            row_liveness = liveness_t.setdefault(target_data['row'], {})
            col_target = target_data['column']
            start, end = col_target, col_source
            if col_target > col_source:
                start, end = col_source, col_target
            for col in xrange(start, end + 1):
                row_liveness.setdefault(col, set()).add(source)
    return liveness, liveness_t


def compute_liveness_phi(layout):
    liveness_columns = {}
    liveness_rows = {}
    for phi in layout.all_phi_nodes:
        phi_row, phi_col = layout.get_cell(phi)
        max_row = min_row = phi_row
        for arg in phi.all_args:
            if arg.is_constant:
                continue
            arg_row, arg_col = layout.get_cell(arg)
            if arg_row > max_row:
                max_row = arg_row
            elif arg_row < min_row:
                min_row = arg_row
            start, end = arg_col, phi_col
            if start > end:
                start, end = end, start
            row_liveness = liveness_rows.setdefault(arg_col, {})
            for col in xrange(start, end + 1):
                row_liveness.setdefault(arg_row, set()).add(phi)
        for row in xrange(min_row, max_row + 1):
            column_liveness = liveness_columns.setdefault(row, {})
            column_liveness.setdefault(phi_col, set()).add(phi)
    # {col: {row: {phi}}}, {row: {col: {phi}}} 
    return liveness_rows, liveness_columns


def allocate_subcolumns(liveness_columns, subcols_present):
    """Allocate each variable to a particular subcolumn.

    This function calculates the number of lines that must travel
    along the column margin, as well as which column and subcolumn
    that each value source provides its value.

    For each source, we calculate its allocated column and subcolumn.

    Additionally, we return the number of subcolumns for each column.
    """
    input_allocation = {} # for each source, the (column, subcolumn)
                           # pairs where its value is broadcasted.
    right_margin = {} # total space required per column.
    for column, column_liveness in liveness_columns.items():
        init_subcol = subcols_present.get(column, 0)
        col_margin, allocations = allocate_crosslinks(column_liveness)
        right_margin[column] = col_margin
        for node, subcolumn in allocations.items():
            input_allocation[node] = (column, init_subcol + subcolumn)
    return input_allocation, right_margin


def allocate_subrows(liveness_rows):
    """Allocate subrows to variables providing values horizontally.

    this section calculates the number of lines that must travel
    along the row margin, as well as which subrow each value source
    should provide its value for each row.

    """
    cross_allocation = {} # for each source, the subrow within each
                           # row that it broadcasts its value on.
    top_margin = {} # total space required per row.
    for row, row_liveness in liveness_rows.items():
        row_margin, allocations = allocate_crosslinks(row_liveness)
        top_margin[row] = row_margin
        for node, subrow in allocations.items():
            cross_allocation.setdefault(node, {})[row] = subrow
    return cross_allocation, top_margin


def compute_standard_links(layout):
    liveness_columns, liveness_rows = compute_liveness(layout)
    col_allocation, col_margin = allocate_subcolumns(liveness_columns, {})
    row_allocation, row_margin = allocate_subrows(liveness_rows)
    return col_allocation, col_margin, row_allocation, row_margin


def compute_phi_links(layout, subcols_present):
    liveness_columns, liveness_rows = compute_liveness_phi(layout)
    col_allocation, col_margin = allocate_subcolumns(liveness_columns,
                                                     subcols_present)
    row_allocation, row_margin = allocate_subrows(liveness_rows)
    return col_allocation, col_margin, row_allocation, row_margin


# each block is laid out like so:
#
# - horizontal input channels -
# constant constant ----\  | | |
# constant constant --\ |  | | | constants truncated to 20 chars
#              _______/ |  | | |
#             /      __/   | | | vertical input channels
#            /      /     /  | |
# opname   [arg] [arg] [arg] | | 
# opname   [proj] [proj]     | |
#            \_____________  | |
#                          \ | |
# - horizontal output (phi) channels -

# if a block has constants, the lines leading to the arguments appear
# in a column on the right hand side of the block.  Otherwise, the
# arguments come straight down from the horizontal input channel, or
# the proj in the block below.

# the client-side layout happens like so:
#
# string lengths are determined.  from these, column widths can be
# determined.  generally, there is five pixels between lines, argument
# boxes are 10 pixels wide, and proj boxes are 15 pixels wide; though
# some proj (eg, EX_TB) may require twice that.
#
# column width is:
#   5 * n_vertical_input_channels +
#   max(len(opname) + 10 * len(args), len(opname) + 15 * len(projs)) +
#   5 * 3 margins
#
# row height is:
#   5 * 2 margins + 20 opname + 16 * n_constants +
#   5 * horizontal_input_chanels + 5 * horizontal_output_channels +
#   (5 * args if needs_rerouting) + 5 * projs
#
# on the server, we determine the nature of the lines, but use
# abstract column and row descriptions to indicate position.  The
# position of each arg box, proj box, and local argument input are fed
# into the line data when laying them out.
#
# the server also determines which args require routing around
# constants, which projs are used, and the direction that the
# connecting arc on the horizontal input channel faces.

CONFIG = {
    "GAP_WIDTH" : 5,
    "ARG_WIDTH" : 16,
    "PROJ_WIDTH" : 16,
    "EST_OP_WIDTH" : 32,
    "OP_HEIGHT" : 32,
    "CONST_HEIGHT" : 16,
    "ARG_HEIGHT" : 16,
    "PROJ_HEIGHT" : 16,
    "MARGIN" : 5
}

def pre_size(layout, node, column_margins, row_margins, config):
    row, col = layout.get_cell(node)
    n_vertical_input = column_margins.get(col, 0)
    n_horizontal_input = row_margins.get(row, 0)
    data = layout.node_data[node]
    nargs = len(node.all_args)
    nproj = len(data.get('proj', ()))
    nconst = sum(1 for arg in node.args if arg.is_constant)
    arg_width = nargs * config['PROJ_WIDTH']
    proj_width = nproj * config['PROJ_WIDTH']
    width = (config['GAP_WIDTH'] * (3 + n_horizontal_input) +
             max(arg_width, proj_width))# + op_width
    height = (config['GAP_WIDTH'] * (2 + n_vertical_input) +
              config['OP_HEIGHT'] +
              (config['CONST_HEIGHT'] + config['GAP_WIDTH']) * nconst +
              config['MARGIN'] * 2)
    data = layout.node_data[node]
    data['base_height'] = height
    data['base_width'] = width

# the distance from the left-hand column to the first arg box is
# 2 * GAP_WIDTH + OP_WIDTH + ARG_WIDTH / 2
# the bottom of the arg line is 
# CONST_HEIGHT * nconst + GAP_WIDTH * n_horizontal_input
# the top of the arg line is
# GAP_WIDTH * valno

# Take 2:
# 
# Things we want to know:
#
# Offset from top of column to first const
#
# Offset to operation
