# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.display.layout import (compute_standard_links,
                                      compute_phi_links, pre_size, CONFIG)


SPECIAL_PROJ = ["V", "M", "EX_V", "EX_TY", "EX_TB", "EX_M"]


def constant_is_special_case(c):
    if isinstance(c.value, int):
        return -9 <= c.value < 100
    return c.value in (None, True, False, ())


def sort_projs(projs):
    sorted_projs = [projs[name] for name in SPECIAL_PROJ if name in projs]
    for name, proj in sorted(projs.items()):
        if name not in SPECIAL_PROJ:
            sorted_projs.append(proj)
    return sorted_projs


def sort_all_projs(nodes):
    all_projs = {}
    for node in nodes:
        if node.is_proj:
            all_projs.setdefault(node.parent, {})[node.proj.name] = node

    sorted_projs = {}
    for node, projs in all_projs.iteritems():
        sorted_projs[node] = sort_projs(projs)
    return sorted_projs


def proj_space(layout, sorted_projs):
    by_row = {}
    for source, projs in sorted_projs.iteritems():
        source_row, _ = layout.get_cell(source)
        value = by_row.get(source_row, -1)
        if len(projs) > value:
            by_row[source_row] = value
    return by_row

def get_data(nodes, layout, config):
    sorted_projs = sort_all_projs(nodes)
    path_number = {}
    vlinks, vmargin, hlinks, hmargin = compute_standard_links(layout)
    vlinks_p, vmargin_p, hlinks_p, hmargin_p = compute_phi_links(layout,
                                                                 vmargin)
    hmargin_proj = proj_space(layout, sorted_projs)

    combined_column_margins = dict(vmargin)
    for col, margin in vmargin_p.iteritems():
        combined_column_margins[col] = (
            combined_column_margins.get(col, 0) + margin)

    combined_row_margins = dict(hmargin)
    for row, margin in hmargin_p.iteritems():
        combined_row_margins[row] = (
            combined_row_margins.get(row, 0) + hmargin_proj.get(row, 0)
            + margin)

    all_node_data = []
    node_data_by_node = {}
    unrendered = {}
    MARGIN = config['MARGIN']
    row_nconsts = {}
    all_args = {}
    arg_id = 0
    for row_number, row in enumerate(layout.data_tabular):
        for col_number, (node, data) in row.iteritems():
            pre_size(layout, node, combined_column_margins,
                     combined_column_margins, config)
            constants = []
            constant_pool = {}
            args = []
            has_phi_arg = False
            for arg in node.mem:
                try:
                    subrow = hlinks[arg][row_number]
                except KeyError:
                    print hlinks.keys()
                    raise
                arg_row, arg_col = layout.get_cell(arg)
                source_col, subcol = vlinks[arg]
                projname = None
                if arg.is_proj:
                    projname = arg.proj.name
                arg_data = {'type' : 'a',
                            'row' : arg_row,
                            'col' : arg_col,
                            'source_col' : source_col,
                            'subrow' : subrow,
                            'subcol' : subcol,
                            'projname' : projname,
                            'path' : [subrow, arg_col, subcol, arg_row,
                            projname],
                            'id' : arg_id}
                args.append(arg_data)
                all_args.setdefault(arg, []).append(arg_data)
                arg_id += 1
            for arg in node.arguments:
                if arg.is_constant:
                    if constant_is_special_case(arg):
                        if isinstance(arg, int):
                            value = str(arg.value)
                        else:
                            value = repr(arg.value)[0]
                        args.append({'type' : 'i', 'value' : value,
                                     'id' : arg_id})
                        arg_id += 1
                    elif arg.value in constant_pool:
                        num = constant_pool[arg.value]
                        args.append({'type' : 'c', 'path' : [num],
                                     'id' : arg_id})
                        arg_id += 1
                    else:
                        num = len(constants)
                        constant_pool[arg.value] = num
                        y_offset = MARGIN + num * (config['CONST_HEIGHT']
                                                   + MARGIN)
                        constants.append({"data" : "%.20r" % (arg.value,),
                                          "y_offset" : y_offset})
                        args.append({'type' : 'c', 'path' : [num],
                                     'id' : arg_id})
                        arg_id += 1
                elif node.is_phi:
                    # phi nodes may only have one non-constant arg
                    if not has_phi_arg:
                        has_phi_arg = True
                        args.append({'type' : 'p', 'path' : [],
                                     'id' : arg_id})
                        arg_id += 1
                else:
                    subrow = hlinks[arg][row_number]
                    arg_row, arg_col = layout.get_cell(arg)
                    source_col, subcol = vlinks[arg]
                    projname = None
                    if arg.is_proj:
                        projname = arg.proj.name
                    arg_data = {'type' : 'a',
                                'row' : arg_row,
                                'col' : arg_col,
                                'source_col' : source_col,
                                'subrow' : subrow,
                                'subcol' : subcol,
                                'projname' : projname,
                                'path' : [subrow, arg_col, subcol, arg_row,
                                projname],
                                'id' : arg_id}
                    args.append(arg_data)
                    all_args.setdefault(arg, []).append(arg_data)
                    arg_id += 1
            row_nconsts[row_number] = max(len(constants),
                                          row_nconsts.get(row_number, 0))

            projs = sorted_projs.get(node, ())
            proj_data = [{"name" : proj.proj.name, "no" : proj.index}
                         for proj in projs]
            node_type = 'phi' if node.is_phi else type(node).__name__
            node_data = {
                'type' : node_type,
                'id' : len(all_node_data),
                'column' : data['column'],
                'row' : data['row'],
                'args' : args,
                'constants' : constants,
                'proj' : proj_data,
            }
            if node.is_op:
                node_data['name'] = node.opname
            node_data_by_node[node] = node_data
            all_node_data.append(node_data)
            for index, proj in enumerate(projs):
                node_data_by_node[proj] = {
                    'proj_no': index,
                    'parent': node_data,
                    'data': proj_data[index]
                }
    height = compute_internal_offsets(all_node_data, hmargin, hmargin_p,
                                      hmargin_proj, config, row_nconsts)

    # Yech, now lets patch up argument paths
    for arg, references in all_args.iteritems():
        for reference in references:
            projname = reference['projname']
            if projname:
                source_data = node_data_by_node[arg.args[0]]
                reference['source_id'] = source_data['id']
                for index, proj in enumerate(source_data['proj']):
                    if proj['name'] == projname.lower():
                        reference['proj_no'] = index

    add_argument_source_data(node_data_by_node, config)

    return {
        'nodes': all_node_data,
        'view_height' : height,
        'vmargin' : combined_column_margins,
        'hmargin' : combined_row_margins,
        'vmargin_p' : vmargin_p,
        'hmargin_p' : hmargin_p,
        'config' : config,
        'row_n_consts' : row_nconsts,
    }


def compute_internal_offsets(all_node_data, hmargin, hmargin_p, hmargin_proj,
                             config, row_nconsts):
    row_start = 0
    row_starts = {}
    row_body = {}
    row_op = {}
    row_phi = {}
    row_proj = {}
    MARGIN = config['MARGIN']
    CONST_HEIGHT = config['CONST_HEIGHT']
    OP_HEIGHT = config['OP_HEIGHT']
    ARG_HEIGHT = config['ARG_HEIGHT']
    for row, nconsts in sorted(row_nconsts.items()):
        row_starts[row] = row_start
        full_height = MARGIN
        full_height += MARGIN * hmargin.get(row, 0)
        row_body[row] = row_start + full_height
        full_height += (MARGIN + CONST_HEIGHT) * nconsts + MARGIN
        row_op[row] = row_start + full_height
        full_height += OP_HEIGHT + MARGIN
        row_proj[row] = row_start + full_height
        full_height += MARGIN * hmargin_proj.get(row, 0) + MARGIN
        row_phi[row] = row_start + full_height
        full_height += MARGIN * hmargin_p.get(row, 0)
        row_start += full_height
    for node in all_node_data:
        row = node['row']
        node['y_hrow_input'] = row_starts[row]
        body = node['y_hrow_body'] = row_body[row] 
        node['y_hrow_proj'] = row_proj[row]
        node['y_hrow_phi'] = row_phi[row]
        op_y = node['y'] = row_op[row]
        for index, constant in enumerate(node['constants']):
            constant['y'] = body + index * (CONST_HEIGHT + MARGIN)
        for index, argument in enumerate(node['args']):
            # TODO: line detail can go here!
            argument['y'] = op_y
        proj_start = row_proj[row]
        for index, proj in enumerate(node['proj']):
            proj['y'] = op_y + ARG_HEIGHT
            proj['out_row'] = proj_start + MARGIN * index
    return row_start


def add_argument_source_data(node_data_by_node, config):
    ARG_HEIGHT = config['ARG_HEIGHT']
    MARGIN = config['MARGIN']
    for node, data in node_data_by_node.iteritems():
        if 'args' not in data:
            continue
        for arg, arg_data in zip(node.args, data['args']):
            if arg_data['type'] == 'a':
                proj_no = None
                source = node_data_by_node[arg]
                if 'parent' in source:
                    proj_no = source['proj_no']
                    source = source['parent']
                arg_data['source_id'] = source['id']
                arg_data['proj_no'] = proj_no
                if proj_no is None:
                    offset = MARGIN
                else:
                    offset = MARGIN * (proj_no + 1)
                arg_data['source_y'] = source['y_hrow_proj'] + offset
            
