# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.base import AnalysisException
from entangled.operation.definitions import Op2


class NullEventHandler(object):
    def handle(self, name, *data):
        pass


class Graph(object):
    def __init__(self, op_types, handler=None):
        self.blocks = []
        self.args = []
        self.all_ops = []
        self.op_types = op_types
        self.backedges = {}
        self.event_handler = handler
        if handler is None:
            self.event_handler = NullEventHandler()
        self.event_handler.handle_ob(self, None, 'graph')
        self.initial_block = self.add_block('initial')
        self.return_block = self.add_block('return')
        self.nonlocal_block = self.add_block('nonlocal')
        self.initial_mem = self.initial_block.op('entangled.initial_memory',
                                                 None, [], [])

    def add_block(self, label, *control):
        block = Block(self, label, *control)
        self.blocks.append(block)
        self.event_handler.handle_ob(block, self, 'ssa-block')
        for source in control:
            self.add_edge(block, source, 'C')
        return block

    def add_edge(self, use, provide, kind):
        if kind != provide.mode:
            raise AnalysisException("Usage of wrong kind?")
        uses = self.backedges.setdefault(provide, [])
        uses.append(use)
        self.event_handler.handle(self, 'edge_' + kind, provide, use)

    def find_problems(self):
        failures = []
        for op in self.all_ops:
            for proj in op.projs:
                uses = self.backedges.get(proj, ())
                if proj.proj.mode == 'C':
                    if len(uses) != 1:
                        failures.append((proj, 'C'))
                elif proj.proj.mode == 'M':
                    if not uses:
                        failures.append(proj, 'M')
        return failures

    def check(self):
        failures = self.find_problems()
        if failures:
            description = []
            for source, mode in failures:
                if mode == 'C':
                    # control flow edges are linear
                    description.append(
                        "%r of mode C is used by %r (should have one target)" %
                        (source, self.backedges.get(source, [])))
                else:
                    # memory edges are relevant
                    description.append(
                        "%r of mode M should have at least one usage" %
                        (source,))
            raise AnalysisException("The following nodes have the incorrect "
                                    "number of uses:\n%s" %
                                    "\n".join(description))


class Block(object):
    def __init__(self, graph, label, *predecessors):
        self.predecessors = list(predecessors)
        if None in self.predecessors:
            raise AnalysisException("No Predecessor?")
        self.operations = []
        self.phi_nodes = []
        self.graph = graph
        self.label = label
        self.exits = []

    def add_predecessor(self, control):
        if control is None:
            raise AnalysisException("No Predecessor?")
        self.predecessors.append(control)
        self.graph.add_edge(self, control, 'C')

    def op(self, name, label, args, mem):
        try:
            ty = self.graph.op_types[name]
        except KeyError:
            raise AnalysisException("No operation: %s" % (name,))
        op = Op2(ty, self, label, mem, args)
        self.graph.event_handler.handle_ob(op, self, 'op',
                                           name, ty, label, mem, args)
        for arg in args:
            self.graph.add_edge(op, arg, 'V')
        for item in mem:
            self.graph.add_edge(op, item, 'M')
        self.graph.all_ops.append(op)
        self.operations.append(op)
        if op.is_phi:
            self.phi_nodes.append(op)
        return op


def on_entangled_initial_memory(operation, state, subs):
    return state.target_graph.initial_memory.dict_projs()


def on_entangled_arg(operation, args, state, subs):
    block = state.target_graph.initial_block
    mem = state.target_graph.initial_memory
    op = block.op('entangled.arg', operation.label, args, [mem])
    return op.dict_projs()


class AnalysisStates(object):
    def __init__(self):
        self.target_states = {}

    def set_state(self, key, value):
        try:
            state = self.target_states[key]
        except KeyError:
            self.target_states[key] = value
        else:
            state.update(value)

class PureState(object):
    def update(self, value):
        raise AnalysisException("Updating a pure state")

class PhiState(object):
    def __init__(self):
        self.sources = []

    def update(self, value):
        self.sources.append(value)


class Analyser(object):
    phi_state_factory = PhiState
    block_state_factory = PhiState
    analysis_state_factory = AnalysisStates

    def __init__(self, handlers):
        self.handlers = handlers

    def analyse(self, source_graph, target_types, *args):
        subs = {}
        op_states = {}
        state = self.analysis_state_factory()
        target_graph = Graph(target_types, source_graph.nargs)
        subs[source_graph.initial_mem] = target_graph.initial_mem
        for source, target in zip(source_graph.args, target_graph.args):
            subs[source] = target

        self._prepare_jump_targets(source_graph, state)

        for block in source_graph.blocks:
            self._analyse_block(target_graph, block, subs, state)

    def _prepare_jump_targets(self, source_graph, state):
        for block in source_graph.blocks:
            block_state = self.block_state_factory()
            state.set_state(block, block_state)
            for index, phi in enumerate(block.phi_nodes):
                phi_state = self.phi_state_factory()
                state.set_state(phi, {'block' : block_state,
                                      'index' : index})

    def _analyse_block(self, target_graph, block, subs, op_states):
        resolved = []
        unresolved = []
        for pred in block.predecessors:
            try:
                resolved.append(subs[pred])
            except KeyError:
                unresolved.append(pred)
            if block is not source_graph.initial_block and not resolved:
                raise AnalysisException("Failed to resolve any "
                                        "predecessors of %r" % (block,))
        # XXX ?
        state = self.merge_states(op_states, resolved, unresolved)
        for op in block.operations:
            if op.is_phi:
                pass
            elif op.is_proj:
                raise AnalysisException("Proj should not be in op list")
            elif op.is_op:
                self._analyse_operation(op, op_states, subs)
            elif op.is_list:
                self._analyse_list(op, op_states, subs)
            elif op.is_constant:
                self._analyse_constant(op, op_states, subs)

    def _analyse_operation(self, op, state, subs):
        mem = [subs[arg] for arg in op.mem]
        op_state = self.merge_states([state[arg] for arg in mem])
        args = [subs[arg] for arg in op.arguments]
        op_projs = op.dict_projs()
        result = self.analyse_operation(op, args, op_state, state)

        # the analysis takes care to preserve existing links -
        # however, new ones can be introduced via the state.
        #
        # note that the results don't have to be a node - 
        for name, proj in op_projs:
            subs[proj] = result[name]
            if proj.proj.mode in ('M', 'C'):
                if proj not in state:
                    raise AnalysisException("Missing state: %r" % (proj,))

    def _analyse_list(self, op, state, subs):
        args = [subs[arg] for arg in op.args]
        result = self.analyse_list(op, args, state)
        subs[op] = result

    def _analyse_constant(self, op, state, subs):
        result = self.analyse_constant(op, state)
        subs[op] = result
        return state

    def analyse_operation(self, op, args, op_state, state):
        handler = self.handlers.get(op.ty.name,
                                    self.default_analyse_operation)
        return handler(self, op, args, op_state, state)

    def default_analyse_operation(self, _, op, args, state):
        return state.block.op(op.ty.name, op.label, args, mem)

    def analyse_list(self, op, args, state):
        return ArgList(args)

    def analyse_constant(self, op, state):
        return op

    def merge_states(self, resolved, unresolved=()):
        pass

    def merge_into_state(self, accum, state):
        pass


class AnalysisStateX(object):
    def jump_to(self, control_edge, target_descriptor, state):
        pass


class AnalyserState(object):
    def __init__(self, graph, analyser):
        self.analyser = analyser
        self.source = graph
        self.work = [(self.analyse_graph, [graph]),
                     (self.create_graph, [graph])]
        self.control = {graph.initial_block : ([], [])} 
        self.analysed = set()
        self.back_edges = {} # TODO: order these by latest update?
        self.analyser.init(self)

    def run_work_one(self):
        f, args = self.work.pop()
        f(self, *args)

    def run_work_complete(self):
        while self.work:
            self.run_work_one()

    def jump_to(self, control_edge, target_descriptor, state):
        if target_descriptor in self.analysed:
            targets = self.back_edges
        else:
            targets = self.control
        edges, states = targets.setdefault(target_descriptor, ([], []))
        edges.append(control_edge)
        states.append(state)

    def create_graph(self, source):
        self.graph = self.analyser.create_graph(source)

    def analyse_graph(self, source):
        blocks = [(self.analyse_block, [block])
                  for block in source.iterblocks()]
        self.work.extend(reversed(blocks))

    def analyse_block(self, block):
        state, operations = self.analyser.analyse_block_pre(self, block)
        self.work.append((self.analyser.analyse_block_post, [state]))
        operations = [(self.analyse_operation, [op, state])
                      for op in block.operations]
        self.work.extend(reversed(operations))


class Analyser3(object):
    def __init__(self, op_types, handlers):
        self.op_types = op_types
        self.handlers = handlers

    def analyse(self, graph):
        state = AnalysisState(graph, self)
        state.run_work_complete()
        return state.graph

    def init(self, state):
        state.frames = {}

    def create_graph(self, source):
        return Graph(self.op_types)

    def analyse_block_pre(self, state, block):
        edges, states = state.control[block]
        if block == state.source.initial_block:
            return state.graph.initial_block
        return state.graph.add_block(block.label, )

    def analyse_block_post(self, state, block_state):
        pass

    def analyse_operation(self, state, op):
        pass


class Analyser2(object):
    def analyse(self, source):
        g = self.create_graph(source)
        state = self.create_initial_state(g)
        targets = [(block, list(block.predecessors))
                   for block in source.iterblocks()]
        # jump target : ([control edges], [states])
        control = {source.initial_block : ([], [state])}
        for sblock in source.iterblocks():
            self.analyse_block(sblock, g, control)

    def analyse_block(self, sblock, g, control):
        input_edges, state = control[sblock]
        block = g.add_block(sblock.label, control[sblock])


class Analyser4(object):
    def analyse_graph(self):
        pass

    def analyse_block(self):
        pass

    def analyse_operation(self):
        pass

    def jump_to(self):
        pass

    def analyse_new_target(self):
        """We are jumping to a target that has not been analysed yet.

        Set up the requisite state so that we can later analyse the
        target.

        """

    def analyse_edge_forward(self):
        """We are jumping to a target that has not been analysed yet.

        Configure the target with the jump data we've been provided.
        """

    def analyse_edge_backward(self):
        """We are jumping to a target that has already been analysed.

        Configure the 
        """
