# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.base import AnalysisException
from entangled.generic.analyser import Graph
from entangled.generic.ianalyser import (IGraphState, IBlockState, IOpState,
                                         IAnalyser)
from entangled.generic.ssi import build_scopes


class BaseAnalyser(IAnalyser):
    def make_graph_state(self, source_graph):
        """Create an IGraphState for analysis.
        """
        return GraphState(self, source_graph)

    def make_block_state(self, graph_state, source_block):
        """Create an IBlockState for later analysis.
        """
        return BlockState(graph_state, source_block)

    def make_op_state(self, block_state, source_op):
        """Create an IOpState state for analysis.
        """
        return OpState(block_state, source_op)

    def analyse_graph(self, source_graph):
        """Analyse the source graph until complete.

        Return the completed IGraphState.
        """
        graph_state = self.make_graph_state(source_graph)
        for block_state in graph_state.blocks():
            self.analyse_block(block_state)
        return graph_state

    def analyse_block(self, block_state):
        """Analyse the IBlockState until complete.
        """
        block_state.initialise()
        for op in block_state.source_block.operations:
            op_state = self.make_op_state(block_state, op)
            self.analyse_op(op_state)

    def analyse_op(self, op_state):
        """Analyse the IOpState.
        """
        op_state.handle()

    def get_handler(self, opname):
        return self.handlers[opname]


class GraphState(IGraphState):
    def __init__(self, analyser, source_graph):
        self.analyser = analyser
        self.source_graph = source_graph
        self.target_graph = Graph()
        self.edge_map = dict((edge, block)
                             for block in source_graph.blocks
                             for edge in block.predecessors)
        self.block_states = {}
        self.work_list = []
        start_block = analyser.make_block_state(self,
                                                source_graph.initial_block)
        self.block_states[start_block.source_block] = start_block
        self.work_list.append(start_block)
        self.ssi_scopes = build_scopes(source_graph.blocks)

    def blocks(self):
        return self.work_list

    def add_jump(self, orig_cond, new_cond, state):
        target = self.edge_map[orig_cond]
        if target in self.block_states:
            state = self.block_states[target]
        else:
            state = self.analyser.make_block_state(self, target)
            self.block_states[target] = state
            self.work_list.append(state)
        state.jump_to(orig_cond, new_cond, state)

    def skip_block(self, orig_target, new_target_cond, transformer):
        """Remove this block from the graph, and forward all jumps.

        :param orig_target: the source-graph block that we are replacing.

        :param new_target: the cond out of the source block that we
                           are populating.

        :param transformer: a function to convert the keys of the
                            binding maps that enter orig_target into
                            binding maps that can work for the new
                            target.

        """

        # get the block associated with the outgoing cond
        new_target_block = self.edge_map[new_target_cond]
        # and the blockstate for the block we're skipping
        block_state = self.block_states[orig_target]
        # check that the block we're skipping is empty
        old_target_block = block_state.target_block
        if any(not op.is_phi for op in old_target_block.operations):
            raise AnalysisError("Removing non-empty block")
        # replace our block_state so that future jumps work
        self.block_states[orig_target] = self.block_states[new_target_block]
        # remove the block from the graph (XXX need a nicer way to do this)
        self.target_graph.blocks.remove(old_target_block)
        # snap the jumps to the new target.
        for orig_cond, new_cond, bindings in block_state.block_entry:
            self.add_jump(new_target_cond, new_cond, transformer(bindings))

class BlockState(IBlockState):

    phi_name = 'entangled.phi'

    def __init__(self, graph_state, source_block):
        self.graph_state = graph_state
        self.source_block = source_block
        self.phi_nodes = []
        self.state = {}
        self.pass_through = {}
        self.block_entry = []
        self.target_block = None

    def initialise(self):
        assert self.target_block is None
        assert not self.state
        self.target_block = self.graph_state.target_graph.block()
        for phi in self.source_block.phi_nodes:
            target_phi = self.make_phi(phi)
            self.state[phi] = target_phi
            self.phi_nodes.append(target_phi)

        _, _, sample_bindings = self.block_entry[0]
        scope = self.graph_state.ssi_scopes.get(self.source_block, ())
        for value in scope:
            if value not in self.pass_through:
                self.pass_through[value] = sample_bindings[value]

        for orig_cond, new_cond, bindings in self.block_entry:
            self._merge(orig_cond, new_cond, bindings)

    def jump_to(self, orig_cond, new_cond, bindings):
        self.block_entry.append((orig_cond, new_cond, bindings))
        if self.target_block is not None:
            self._merge(orig_cond, new_cond, bindings)

    def _merge(self, orig_cond, new_cond, bindings):
        self.target_block.add_predecessor(new_cond)
        source_block_index = self.source_block.predecessors.index(orig_cond)
        for source, target in zip(self.source_block.phi_nodes,
                                  self.phi_nodes):
            source_value = source.arguments[source_block_index]
            self.merge_phi(target, bindings[source_value])
        self._sanity_check_bindings(bindings)

    def _sanity_check_bindings(self, bindings):
        missing_values = set(self.pass_through) - set(bindings)
        if missing_values:
            report = '\n'.join(map(repr, sorted(missing_values)))
            raise AnalysisException("Missing bindings for some required "
                                    "values:\n" + report)

        for value, binding in self.pass_through.iteritems():
            if value not in bindings:
                raise AnalysisException("Missing binding for: %r" % (value,))
            if bindings[value] != binding:
                raise AnalysisException(
                    "Different values provided for: %r\n orig: %r\n next: %r"
                    % (value, binding, bindings[value]))

    def make_phi(self, source_phi):
        return self.target_block.op('entangled.phi', source_phi.offset, [], [])

    def merge_phi(self, target, binding):
        target.arguments.append(binding)

    def get(self, key, default=None):
        if key in self.state:
            return self.state[key]
        return self.pass_through(key, default)


class OpState(IOpState):
    def __init__(self, block_state, source_op):
        self.block_state = block_state
        self.source_op = source_op
        self.analyser = self.block_state.graph_state.analyser

    @property
    def graph_state(self):
        return self.block_state.graph_state

    @property
    def block(self):
        return self.block_state.target_block

    def get_offset(self):
        return self.source_op.offset

    def op(self, name, mems, arguments):
        return self.block.op(name, self.get_offset(), mems, arguments)

    def handle(self):
        sop = self.source_op
        memory = [self.block_state.get(mem) for mem in sop.mem]
        arguments = [self.block_state.get(arg) for arg in sop.arguments]
        handler = self.analyser.get_handler(sop.opname)
        handler(self, memory, *arguments)
