# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Additional analyser classes that honestly belong elsewhere.

"""

from entangled.base import AnalysisException
from entangled.generic.analyser import Graph


MISSING = object()

def _get_option(provided, default):
    if provided is MISSING:
        return default
    return provided


class FrameState(object):
    def __init__(self, block_state, stack, local, exc_info, result,
                 mem=None, block_stack=None):
        self.block_state = block_state
        self.stack = stack
        self.local = local
        self.exc_info = exc_info
        self.result = result
        self.mem = mem
        self.block_stack = block_stack

    @property
    def analyser(self):
        return self.block_state.analyser

    def set_to(self, block_state=MISSING, stack=MISSING, local=MISSING,
               exc_info=MISSING, result=MISSING, mem=MISSING):
        return FrameState(
            _get_option(block_state, self.block_state),
            _get_option(stack, self.stack),
            _get_option(local, self.local),
            _get_option(exc_info, self.exc_info),
            _get_option(result, self.result),
            _get_option(mem, self.mem),
        )


# This may be a more sensible idea than the SSAFrameState stuff.
#
# Currently unused.
class FrameStatePhi(FrameState):
    def _make_phi(self, ty, offset, args):
        # TODO: mode?
        return self.block.op('entangled.phi.%s' % ty, offset, args, [])

    def __init__(self, block_state):
        offset = block_state.source_block.offset
        nlocals = block_state.graph_state.nlocals
        local = []
        for index in xrange(nlocals):
            local.append(self._make_phi('local', offset, [Constant(index)]))
        exc_info = [
            self._make_phi('exc_ty', offset, []),
            self._make_phi('exc_v', offset, []),
            self._make_phi('exc_tb', offset, []),
        ]
        result = self._make_phi('result', offset, [])
        mem = self._make_phi('mem', offset, [])
        super(FrameStatePhi, self).__init__(
            Stack(), local, exc_info, result, mem
        )

    def merge(self, cond, target):
        self.block.add_predecessor(cond)
        self.mem.arguments.append(target.mem)
        self.result.arguments.append(target.result)
        for loc, phi in zip(target.local, self.local):
            phi.arguments.append(loc)
        for exc, phi in zip(target.exc_info, self.exc_info):
            phi.arguments.append(exc)
        self.stack.merge(target.stack)
