# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Interfaces for the generic graph analysis/transform.

Jokes about optometrists are welcome.
"""

class IGraphState(object):
    def get_initial_state(self):
        """Get the analysis-specific IBlockState for the start block.
        """

    def blocks(self):
        """Iterable of valid IBlockStates for analysis.

        We guarantee that the IBlockState yielded is ready - that is,
        it has at least one predecessor; and so has all the bindings
        set-up for analysis.

        """

    def add_jump(self, orig_cond, new_cond, state):
        """Take a jump (orig_cond) in the original graph.

        When interpreting an operation with control flow, it could be
        that several generated operations may cause the jump.  The
        add_jump method can be called several times with the same
        orig_cond - this creates several jumps in the target graph.
        Alternatively, you could never call it, which will prevent the
        jump following the orig_cond from being represented in the
        final graph.

        orig_cond: The original projection representing the jump.

        new_cond: A projection of an operation in the new graph that
                  leads to following the jump.

        state: A dictionary mapping projections in the original graph
               with values that make sense in the analysis.  Usually
               this will be projections in the target graph, but they
               could also be bundled with data relevant to the
               analysis.

        """


class IBlockState(object):
    def jump_to(self, orig_cond, new_cond, bindings):
        """Record a jump to this block.

        If the block has not been analysed yet, we'll simply record
        the jump to use when initialising later.

        Otherwise, set up our phi nodes from the bindings and apply
        the new_cond to the target block.

        """

    def initialise(self):
        """Initialise the block state from known predecessors.
        """

    def make_phi(self, source_phi):
        """Create a target binding for the source phi operation.
        """

    def merge_phi(self, target_phi, binding):
        """Add the binding to a phi node binding that we've generated.
        """

    def get(self, key, default=None):
        """Get the value associated with the given source node.

        When analysing an operation, we usually want values in the
        target graph to use as parameters.  The analysis will use this
        function to get values relevant to the analysis rather than
        source nodes.

        In particular, this function is invoked on constants.  If you
        want to transform constants, implement their special handling
        here.

        """


class IOpState(object):
    """Analysis information relevant for a single operation.

    The analyser as it exists uses the IOpState as a functional
    abstraction for the most part - that is, there is only one live at
    any time - so you are free to create only one and re-use it from
    operation to operation.

    attribute: graph_state - the related graph_state
    attribute: block_state - the related block_state

    """

    def get_offset(self):
        """Get the debugging offset of this operation.

        This will often include the bytecode address of the original
        operation, which makes it easy to relate target operations
        back to source operations.

        """

    def op(self, name, mems, arguments):
        """Create an operation in the source graph.
        """

    def handle(self):
        """Analyse this operation.
        """


class IAnalyser(object):
    """Interface that an analysis is expected to provide.

    :attr handlers: a map from opname to handler function. For each
                    operation the handler will be extracted and called
                    with the IOpState, the memory bindings, and the
                    arguments bindings for the operation.  If the
                    opname is not found, a KeyError is raised, if that
                    is not desired, the get_handler method can be
                    supplied instead.

    """

    def make_graph_state(self):
        """Create an IGraphState for analysis.
        """

    def make_block_state(self, graph_state, source_block):
        """Create an IBlockState for later analysis.
        """

    def make_op_state(self, block_state, source_op):
        """Create an IOpState state for analysis.
        """

    def analyse_graph(self, source_graph):
        """Analyse the source graph until complete.

        Return the completed IGraphState.
        """

    def analyse_block(self, block_state):
        """Analyse the IBlockState until complete.
        """

    def analyse_op(self, op_state):
        """Analyse the IOpState.
        """

    def get_handler(self, opname):
        """Look-up a handler for this operation type.
        """
