# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
from entangled.base import AnalysisException
from entangled.operation.definitions import Block, SSAElement

class RenderContext(object):
    def __init__(self):
        self.names = {}
        self.name_count = {}

    def output(self, value):
        print(value)

    def name(self, value):
        try:
            return self.names[value]
        except KeyError:
            name = value.__class__.__name__
            count = self.name_count.get(name, 0)
            self.name_count[name] = count + 1
            name = '%s_%s' % (name, count)
            self.names[value] = name
            return name

    def show_graph(self, graph, name):
        self.output("function %s:" % (name,))
        blocks = list(graph.blocks)
        blocks.sort(key=lambda block: block.operations[0].offset)
        for block in blocks:
            self.show_block(block)

    def show_block(self, block):
        self.output("  block %s:" % (self.name(block),))
        if not block.predecessors:
            self.output('    no_predecessors')
        for index, predecessor in enumerate(block.predecessors):
            pred_name = self.name(predecessor.block)
            self.output('    pred (%s).%s:' % (
                pred_name, self.show_ssa_element_value(predecessor)))
            for phi in block.phi_nodes:
                arguments = phi.arguments if phi.arguments else phi.mem
                prev = "BAD_PHI"
                if index < len(arguments):
                    prev = self.show_ssa_element_value(arguments[index])
                self.output('      %s = %s(%s)' % (
                    self.name(phi), phi.ty.name, prev))
        for index, operation in enumerate(block.operations):
            if operation.is_phi:
                continue
            elif operation.is_op:
                self.output('    %s' % (
                    self.show_ssa_element_definition(operation),))

    def show_ssa_element_values(self, elements):
        return ', '.join(self.show_ssa_element_value(element)
                         for element in elements)

    def show_ssa_element_value(self, element):
        if element.is_phi:
            return self.name(element)
        elif element.is_proj:
            return '%s.%s' % (self.name(element.parent), element.proj.name)
        elif element.is_constant:
            return repr(element)
        elif element.is_op:
            raise AnalysisException("op != value for the purpose of show")
        elif element.is_list:
            return '[%s]' % (self.show_ssa_element_values(element.args),)
        raise AnalysisException("unknown value type: %r" % element)

    def show_ssa_element_definition(self, element):
        if element.is_phi:
            return AnalysisException("phi need to be defined with the block")
        if element.is_op:
            return '%s = %s<%s>(%s, %s)' % (
                self.name(element),
                element.ty.name,
                element.offset,
                self.show_ssa_element_values(element.arguments),
                self.show_ssa_element_values(element.mem),
            )
        return AnalysisException("defining non-op: %r" % (element,))
