# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

def build_scopes(blocks):
    """Collect (SSI style) usage information per source block.

    The values that may be used within a given block include:

    * constants.

    * values produced by operations within the block can be used by
      later operations within the block.

    * phi nodes bound by the block.  these typically have different
      values depending on which predecessor passed control to this
      block.

    * values produced by preceeding blocks.  for these to work without
      a phi node, they must have exactly the same value in all of the
      block's predecessors.

    It's the last class of values that we focus on here.  Our
    applicative analyser expects to pass a newly produced value for
    use in analysing subsequent operations.  If our graphs were in SSI
    form this would be obvious; but we can only assume SSA (often
    minimal SSA), so we calculate operation liveness per block.

    This information cannot really be used for generating
    garbage-collection information; in particular it's possible for
    programs to rely on the presence of an object on the expression
    stack that is nevertheless unused by the end of the expression to
    prevent its being finalised until the expression is complete.
    It's up to your own compiler passes to discover language-defined
    liveness information.

    """

    live_values = {}
    unanalysed_dependencies = {}
    # values_produced = {}
    def add_dep(source, value):
        live_values.setdefault(source, set()).add(value)
        if source != value.block:
            unanalysed_dependencies.setdefault(source, []).append(arg)

    # right now, we're using the `source != value.block` test above
    # for block in blocks:
    #     values_produced[block] = produced = set()
    #     for op in block.operations:
    #         produced.add(op)
    for block in blocks:
        for op in block.operations:
            for arg in op.arguments:
                # two cases we care about: it's a phi node for this
                # block, or it's a node that comes from another block.
                if arg.is_phi and arg.block == block:
                    for control, value in zip(block.predecessors,
                                              arg.arguments):
                        add_dep(control.block, value)
                elif arg.is_op and arg.block != block:
                    for source in block.predecessors:
                        add_dep(source.block, arg)
    while unanalysed_dependencies:
        block, values = unanalysed_dependencies.popitem()
        for value in values:
            for pred in block.predecessors:
                add_dep(pred.block, value)
    return live_values
