# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.op_types import phi_types

DEFAULT_MEM = object()

class Block(object):
    def __init__(self, types, number):
        self.types = types
        self.number = number
        self.phi_nodes = []
        self.exits = set()
        self.operations = []
        self.predecessors = []

    def add_op(self, op):
        self.operations.append(op)
        if op.is_phi:
            self.phi_nodes.append(op)

    def add_predecessor(self, block):
        self.predecessors.append(block)

    def op(self, name, offset, args, mem=DEFAULT_MEM):
        op_type = self.types[name]
        if mem is DEFAULT_MEM:
            mem = self.mem
        if op_type.stararg:
            if len(args) < len(op_type.args):
                print op_type.name, op_type.args, len(args)
        else:
            if len(args) != len(op_type.args):
                print op_type.name, op_type.args, len(args)
        op = Op2(op_type, self, offset, mem, args)
        self.operations.append(op)
        return op


class BaseBlockBuilder(object):
    def __init__(self):
        self.blocks = {}


class SSAElement(object):
    is_op = False
    is_constant = False
    is_list = False
    is_phi = False
    is_proj = False


class Proj(SSAElement):
    is_op = True
    is_proj = True

    def __init__(self, parent, proj, index):
        self.parent = parent
        self.proj = proj
        self.index = index

    @property
    def mode(self):
        return self.proj.mode

    @property
    def block(self):
        return self.parent.block

    def get_args(self):
        return [self.parent]

    arguments = args = all_args = property(get_args)

    @property
    def mem(self):
        return ()

    @property
    def name(self):
        return "%s.%s" % (self.parent.name, self.proj.name)

    @property
    def offset(self):
        return self.parent.offset

    def pre_clone(self, block=None, mem=None, args=None):
        assert len(args) == 1, repr(args)
        return Proj(args[0], self.proj, self.index)

    def __repr__(self):
        return '%s[%r]' % (self.proj.name, self.parent)


class Op2(SSAElement):
    is_op = True
    is_proj = False

    def __init__(self, ty, block, offset, mem, args):
        self.ty = ty
        self.block = block
        self.offset = offset
        self.mem = mem
        self.arguments = args
        self.proj = {}
        self.projs = []
        for arg, arg_ty in zip(args, ty.args):
            setattr(self, 'arg_' + arg_ty, arg)
        if ty.stararg:
            setattr(self, 'arg_' + ty.stararg, args[len(ty.args):])
        for index, proj_ty in enumerate(ty.projs):
            proj = Proj(self, proj_ty, index)
            self.projs.append(proj)
            self.proj[proj_ty.name] = proj
            setattr(self, 'proj_' + proj_ty.name, proj)

    def pre_clone(self, block=None, mem=None, args=None):
        return Op2(self.ty, block or self.block, self.offset,
                   mem or [], args or [])

    @property
    def name(self):
        return self.ty.name

    @property
    def all_args(self):
        "Return all dependencies, memory and other arguments"
        return self.mem + self.arguments

    @property
    def args(self):
        "Deprecated"
        return self.mem + self.arguments

    @property
    def is_phi(self):
        return self.ty.is_phi

    @property
    def opname(self):
        return self.ty.name

    def sorted_proj(self):
        return list(self.projs)

    def dict_projs(self):
        return dict((proj.proj.name, proj) for proj in self.projs)

    def __getattr__(self, name):
        raise AttributeError("%r has no attribute %s" % (self, name))

    def __repr__(self):
        return "%s<%s>/%d" % (self.ty.name, self.offset, len(self.arguments))
        

class Op(SSAElement):
    is_op = True
    def __init__(self, name, block, offset, args):
        assert isinstance(name, str)
        self.name = name
        self.block = block
        self.offset = offset
        self.args = args
        self._proj = {}
        block.add_op(self)

    def pre_clone(self, block=None, args=None):
        return Op(self.name, block or self.block, self.offset, args or [])

    @property
    def is_phi(self):
        return self.name.startswith('entangled.phi')

    @property
    def args_flat(self):
        for arg in self.args:
            if isinstance(arg, list):
                for subarg in arg:
                    yield subarg
            else:
                yield arg

    @property
    def opname(self):
        return self.name

    def __lt__(self, other):
        return (self.offset, self.block, self.name, self.args) < other

    def __ge__(self, other):
        return (self.offset, self.block, self.name, self.args) >= other

    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, Op):
            return NotImplemented
        if (self.offset != other.offset or self.block != other.block or
            self.name != other.name):
            return False
        if len(self.args) != len(other.args):
            return False
        return all(arg is other_arg
                   for arg, other_arg in zip(self.args, other.args))

    def __hash__(self):
        return hash((self.offset, self.name))

    def sorted_proj(self):
        result = []
        for name in SPECIAL_PROJ:
            if name in self._proj:
                result.append(self._proj[name])
        remaining = []
        for name, proj in self._proj.itervalues():
            if name not in SPECIAL_PROJ:
                remaining.append(proj)
        remaining.sort(key=lambda proj: proj.opname)
        return result + remaining

    def __repr__(self):
        return "%s<%s>/%d" % (self.name, self.offset, len(self.args))

    def __getattr__(self, name):
        if name.startswith('proj_'):
            op = Op('%s.%s' % (self.name, name), self.block, self.offset,
                    [self])
            self._proj[name[5:]] = op
            setattr(self, name, op)
            return op
        super(Op, self).__getattribute__(name)


class Constant(SSAElement):
    is_constant = True
    mode = 'V'
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "<%r>" % (self.value,)

    def __eq__(self, other):
        if isinstance(other, Constant):
            return self.value == other.value
        return NotImplemented

    def __hash__(self):
        return hash((self.value,))


class ArgList(SSAElement):
    is_list = True
    def __init__(self, values):
        self.args = values

    @property
    def all_args(self):
        return self.args

    @property
    def mem(self):
        return ()

    @property
    def arguments(self):
        return self.args

    def __repr__(self):
        return "ArgList(%s)" % ', '.join(repr(arg) for arg in self.args)


