# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.base import AnalysisException
from entangled.operation.definitions import Constant
from entangled.op_types import op_definer, op_language, C, V, M, ProjType


def _mkproj(handler, modes):
    return [ProjType('%s_%s' % (handler, mode), mode) for mode in modes]


ALT_PROJ = _mkproj('alt', 'CVM')
EXIT_PROJ = _mkproj('exit', 'CVM')
SIZE_PROJ = ProjType('stack_size', 'V')
SHOW_BLOCK_STACK = True


handler_proj = {
    'raise' : _mkproj('raise', 'CVM'),
    'return' : _mkproj('return', 'CVM'),
    'break' : _mkproj('break', 'CVM'),
#    'reraise' : _mkproj('reraise', 'CVM'),
#    'control_loop' : _mkproj('control_loop', 'CVM'),
    'continue' : _mkproj('continue', 'CVM'),
}


BLOCK_MATCH = {
    'except' : ['raise'],
    'finally' : ['raise', 'return', 'break', 'continue'],
    'loop' : ['break', 'continue'],
    'return' : ['return'],
    'restore': [],
}


def get_matching_targets(block_stack, event, expected_events):
    targets = {}
    events = [event]
    if event == 'reraise':
        events = expected_events
    for event in events:
        for target, block_type, _ in reversed(block_stack):
            if event in BLOCK_MATCH[block_type]:
                targets[event] = (target, block_type)
                break
    return targets


def manage_exit_types(ty, handler_stack, expected_handlers):
    """Track pseudo-exceptions that can reach a given block.

    Return the list of outgoing edge types this op can raise.
    """
    # special handling for 'continue': we don't want to lose the jump
    # offset, and we want the END_FINALLY operation to know that it
    # could be unrolling a break or continue event.
    if ty.name in ('continue', 'break'):
        for handler in reversed(handler_stack):
            target_offset, block_type, stack_size = handler
            if block_type == 'loop':
                expected_handlers.setdefault(
                    target_offset, set()
                ).add((ty.name, oparg))
                break
            if block_type == 'finally':
                expected_handlers.setdefault(
                    target_offset, set()
                ).add((ty.name, oparg))
    elif ty.name in ('return', 'raise'):
        for handler in reversed(handler_stack):
            target_offset, block_type, stack_size = handler
            if block_type == 'finally':
                expected_handlers.setdefault(
                    target_offset, set()
                ).add((ty.name, oparg))

    # if this is an END_FINALLY we need to collect the causes
    if ty.name == 'reraise':
        reraise_causes = set()
        target_offset, block_type, _ = handler_stack[-1]
        assert block_type == 'restore', repr(handler_stack)
        for cause, _ in expected_handlers.get(target_offset, ()):
            reraise_causes.add(cause)
        return sorted(reraise_causes)
    return ty.outgoing_edges


def build_opset(config):
    """Build the language that trivial_ssa will output.
    """
    op = op_definer("bytecode", [])
    operations = []
    for name, operation in config.operations.iteritems():
        proj = []
        if not operation.noreturn:
            proj.extend([M, C, V])
        if operation.block_type:
            if not operation.no_jump:
                proj.extend(EXIT_PROJ)
            proj.append(SIZE_PROJ)
        elif operation.hasjabs or operation.hasjrel:
            proj.extend(ALT_PROJ)
        edges = set()
        for ty in operation.nonlocal_types:
            edges.update(ty.outgoing_edges)
        for edge in sorted(edges):
            proj.extend(handler_proj[edge])
        operations.append(op("bytecode." + name, ['state', 'opcode'], proj,
                             source_type=operation))
    for block_type in ['except', 'loop', 'finally', 'return']:
        operations.append(op('landing.' + block_type, ['state'], [M, C, V]))
    return operations


def check_compatible(provided, new):
    if provided == new or (provided and provided[:-1] == new and
                           provided[-1][1] == 'restore'):
        return
    if provided:
        existing, (_, reasons) = provided[:-1], provided[-1]
        if reasons == 'restore':
            return
    raise AnalysisException("non-compatible block stacks:\n %r\n %r" %
                            (provided, new))


# general note: this function walks through the bytecode - in order -
# wiring up jumps and maintaining the block-stack so as to determine
# the control flow that is possible within the code.
#
# what it does not set up is the outgoing edges from operations that
# reraise.


def trivial_ssa(graph, operations, labels):
    # a map from bytecode offset to the block and its phi nodes
    block_headers = {}

    # blockstack changes from nonlocal jumps - a map from bytecode
    # offset to blockstack.
    prepared_blockstack = {}

    all_loop_headers = []
    landings = {}

    def _create_jump_target(offset_bc, block=None):
        if block is None:
            block = graph.add_block(offset_bc)
        # XXX do these have proj?
        phi_V = block.op('entangled.phi', offset_bc, [], [])
        phi_M = block.op('entangled.phi.mem', offset_bc, [], [])
        block_headers[offset_bc] = block, phi_V, phi_M

    # for offset_bc in labels:
    #     _create_jump_target(graph, offset_bc)
    _create_jump_target('raise', graph.nonlocal_block)
    _create_jump_target('return', graph.return_block)

    def jump(offset_bc, control, value, mem):
        try:
            block, phi_V, phi_M = block_headers[offset_bc]
        except KeyError:
            _create_jump_target(offset_bc)
            block, phi_V, phi_M = block_headers[offset_bc]
        if control is not None:
            block.add_predecessor(control)
            phi_V.arguments.append(value)
            phi_M.arguments.append(mem)
        return block, phi_V.proj_V, phi_M.proj_M

    def land(offset_bc, reason, block_type, control, value, mem):
        """Jump to a handler, with the specified reason.

        This is like the regular jump, but also produces a
        reason-specific operation in-between.  This gives an
        opportunity to set up the frame appropriately in the unstack
        phase.

        """
        # produce a jump to the landing
        offset = (offset_bc, reason)
        block, V, M = jump(offset, control, value, mem)
        if (offset_bc, reason) not in landings:
            op = block.op('landing.' + block_type, offset, [V], [prev_M])
            landings[offset_bc, reason] = op
            if reason != 'continue':
                jump(offset_bc, op.proj_C, op.proj_V, op.proj_M)

    # The handler_stack is a stack of (offset, {reason}) pairs.  These can
    # indicate where to jump if a matching nonlocal type is encountered.
    handler_stack = [
        ('except', 'except', None),
        ('return', 'return', None),
    ]
    # this is a stack of blocks currently being handled - it
    # represents the unrollers present on the stack.
    handling = []
    previous = graph.initial_block.op('entangled.start', -1, [],
                                      [graph.initial_mem.proj_M])
    prev_M = previous.proj_M
    prev_C = previous.proj_C # 'start' op
    prev_V = previous.proj_V
    noreturn = True
    current_block = None

    expected_handlers = {}

    for bop in operations:
        yield bop
        if SHOW_BLOCK_STACK:
            print "%s %s" % (bop.opname,
                             '.'.join("%s(%s)" % (block_type, offset)
                                      for offset, block_type, _
                                      in handler_stack))
        # if we've jumped here via an edge that removes a block, pop
        # the blockstack now
        if bop.offset_bc in prepared_blockstack:
            if not noreturn:
                check_compatible(prepared_blockstack[bop.offset_bc],
                                 handler_stack)
            handler_stack = prepared_blockstack[bop.offset_bc]

        # if it's possible to jump to this operation, setup the phi
        # nodes as the predecessors, otherwise create the block for
        # this op
        if bop.offset_bc in labels:
            block, prev_V, prev_M = jump(bop.offset_bc,
                                         prev_C, prev_V, prev_M)
        else:
            block = graph.add_block(bop.offset_bc, prev_C)

        # create the op
        opname = 'bytecode.%s' % bop.opname
        opcode = Constant(bop.oparg)
        op = block.op(opname, bop.offset_bc, [prev_V, opcode], [prev_M])
        conf = op.ty.source_type

        # if bop.opname == "CALL_FUNCTION":
        #     import pdb
        #     pdb.set_trace()

        # if the op raises/continues/breaks/returns etc, connect this
        # operation to the handler
        assert not conf.nonlocal_types or len(conf.nonlocal_types) == 1
        if conf.nonlocal_types:
            ty, = conf.nonlocal_types
            exit_types = manage_exit_types(ty, handler_stack,
                                           expected_handlers)
            targets = get_matching_targets(handler_stack, ty.name, exit_types)
            for event, (target, block_type) in targets.iteritems():
                C, V, M = [op.proj['%s_%s' % (event, mode)] for mode in 'CVM']
                land(target, ty.name, block_type, C, V, M)

        target = bop.jump_target
        if target is not None:
            if conf.block_type:
                block_type_target = None
                try:
                    block_type_cont, block_type_target = conf.block_type
                except ValueError:
                    block_type_cont, = conf.block_type

                # this operation is beginning a python VM block -
                # perhaps a try block or loop. add it to the block stack.

                # first, record the expected handler stack at the target
                target_handler_stack = list(handler_stack)
                if block_type_target is not None:
                    target_handler_stack.append((target, block_type_target,
                                                 op.proj_stack_size))
                if target in prepared_blockstack:
                    check_compatable(prepared_blockstack[target],
                                     target_handler_stack)
                else:
                    prepared_blockstack[target] = target_handler_stack

                # add the default handler
                if block_type_cont is not None:
                    handler_stack = handler_stack + [
                        (target, block_type_cont, op.proj_stack_size)
                    ]

                if not conf.no_jump:
                    #if 'loop' in conf.block_type:
                    jump(target,
                         op.proj_exit_C,
                         op.proj_exit_V,
                         op.proj_exit_M)
                    all_loop_headers.append(bop.offset_bc)
            elif any(ty.name == 'continue' for ty in conf.nonlocal_types):
                # in this case, the jump is from the *loop header* to
                # the target.
                for loop_header, block_type, _ in reversed(handler_stack):
                    if block_type == 'loop':
                        break
                else:
                    raise AnalysisException("continue outside of loop block")
                jump(target, op.proj_alt_C, op.proj_alt_V, op.proj_alt_M)
            else:
                # XXX need to handle `continue`
                # the operation itself performs the jump
                jump(target, op.proj_alt_C, op.proj_alt_V, op.proj_alt_M)

        if conf.pops_block:
            if len(handler_stack) <= 2:
                raise AnalysisException("bad op.pops_block from empty stack: %s" % (conf,))
            handler_stack = handler_stack[:-1]
            current_block = None

        noreturn = conf.noreturn
        if conf.noreturn:
            prev_C = prev_V = prev_M = None
        else:
            prev_C = op.proj_C
            prev_V = op.proj_V
            prev_M = op.proj_M

    # we've considered all operations - final checks & fixes
    if len(handler_stack) != 2:
        raise AnalysisException(
            "function ended with non-empty block stack: %r" %
            (handler_stack[2:],)
        )

    for offset in all_loop_headers:
        backward_jumps = expected_handlers.get(target_offset, ())
        continues = [target for cause, target in backward_jumps
                     if cause == 'continue']
        if not continues:
            continue # no continues to process here

        op = landings[offset, 'continue']
        proj_C, proj_V, proj_M = op.proj_C, op.proj_V, op.proj_M
        block = op.block
        for target in continues:
            block = graph.add_block('continue_%s' % target)
            block.add_predecessor(proj_C)
            op = block.op('continue_match', offset, [proj_V, target], [proj_M])
            jump(target, op.proj_C, proj_V, proj_M)
            proj_C = op.proj_alt_C
        # TODO: connect remaining proj_C to a BAD block.
