# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import dis
import attr

STANDARD_NAMES = ("hascompare hasconst hasfree hasjabs hasjrel haslocal "
                  "hasname hasnargs").split()

CONFIG_NAMES = STANDARD_NAMES + [
    "noreturn", "hasarg", "pops_block", "no_jump",
]

CONFIG_NAMES_DEFAULT = {
}

STANDARD_NO_RETURN = (
    'BREAK_LOOP', 'CONTINUE_LOOP', 'JUMP_ABSOLUTE', 'JUMP_FORWARD',
    'RAISE_VARARGS', 'RETURN_VALUE', 'STOP_CODE'
)

CONFIG_UNABBREV = {
    "cmp" : "hascompare",
    "const" : "hasconst",
    "free" : "hasfree",
    "jabs" : "hasjabs",
    "jrel" : "hasjrel",
    "local" : "haslocal",
    "name" : "hasname",
    "nargs" : "hasnargs",
    "no_return" : "noreturn",
    "oparg" : "hasarg",
    "popblock" : "pops_block",
    "nojump" : "no_jump"
}

CONFIG_ABBREV = dict((v, k) for k, v in CONFIG_UNABBREV.items())


OPARG_REQD = set("hasjabs hasjrel hasconst hasname hasfree hascompare "
                 "haslocal hasnargs".
                 split())


@attr.s
class NonlocalType(object):
    name = attr.ib()
    handlers = attr.ib(repr=False)
    outgoing_edges = attr.ib(repr=False)


# NONLOCAL_TYPES = {
#     'raise' : MAY_RAISE,
#     'return' : MAY_RETURN,
#     'control_loop' : MAY_BREAK,
# }


class OperationType(object):
    def __init__(self, number, name, options, nonlocal_types, block_type):
        self.number = number
        self.name = name
        self.nonlocal_types = nonlocal_types
        self.block_type = block_type
        for config_name in CONFIG_NAMES:
            setattr(self, config_name, options.get(config_name, False))


class Config(object):

    MAY_RAISE = NonlocalType('raise', ['except', 'finally'], ['raise'])
    MAY_RERAISE = NonlocalType('reraise', ['except', 'finally', 'loop',
                                           'return'],
                               ['raise', 'break', 'return', 'continue'])
    MAY_RETURN = NonlocalType('return', ['finally', 'return'], ['return'])
    MAY_BREAK = NonlocalType('break', ['finally', 'loop'], ['break'])
    MAY_CONTINUE = NonlocalType('continue', ['finally', 'loop'], ['continue'])

    def __init__(self, implementation, major, minor):
        self.implementation = implementation
        self.major = major
        self.minor = minor
        self.opname = self.opnames = [None] * 256
        self.opmap = {}
        self.operations = {}
        self.hasarg = []
        self.hascompare = []
        self.hasconst = []
        self.hasfree = []
        self.hasjabs = []
        self.hasjrel = []
        self.haslocal = []
        self.hasname = []
        self.hasnargs = []
        self.noreturn = []
        self.no_jump = []
        self.pops_block = []
        self.default_nonlocal_types = [self.MAY_RAISE]

    @property
    def no_return(self):
        return self.noreturn

    @property
    def name(self):
        return '%s-%s.%s' % (self.implementation.lower(),
                             self.major, self.minor)

    def has_argument(self, opno):
        return self.operations[self.opnames[opno]].hasarg

    def define(self, number, name, flags=(), nonlocal_types=None,
               block_type=None):
        self.opnames[number] = name
        self.opmap[name] = number
        options = {}
        for flag in flags:
            option = CONFIG_UNABBREV[flag]
            options[option] = True
            getattr(self, option).append(number)
        if any(opt in OPARG_REQD for opt in options):
            options["hasarg"] = True
        if nonlocal_types is None:
            nonlocal_types = self.default_nonlocal_types
        self.operations[name] = OperationType(number, name, options,
                                              nonlocal_types, block_type)


class NativeConfig(object):
    def __init__(self, opname, opmap, hascompare=(), hasconst=(),
                 hasfree=(), hasjabs=(), hasjrel=(), haslocal=(),
                 hasname=(), hasnargs=(), no_return=(), have_argument=None):
        self.opname = opname
        self.opmap = opmap
        self.hascompare = hascompare
        self.hasconst = hasconst
        self.hasfree = hasfree
        self.hasjabs = hasjabs
        self.hasjrel = hasjrel
        self.haslocal = haslocal
        self.hasname = hasname
        self.hasnargs = hasnargs
        self.no_return = no_return
        self.HAVE_ARGUMENT = have_argument
        self.operations = {}

    def get_options(self, opnum):
        options = []
        for option in STANDARD_NAMES:
            oplist = getattr(self, option, {})
            if opnum in oplist:
                options.append(option)
        if opnum > self.HAVE_ARGUMENT and not options:
            options.append("hasarg")
        if opnum in self.no_return:
            options.append("noreturn")
        return options

    def generate(self):
        ops = sorted((name, index) for index, name in enumerate(self.opname)
                     if '<' not in name)
        for opname, index in ops:
            options = self.get_options(index)
            yield index, opname, [CONFIG_ABBREV[option]
                                  for option in options]

    @classmethod
    def from_native(cls, src=dis):
        no_return = [src.opmap[name] for name in STANDARD_NO_RETURN
                     if name in src.opmap]
        result = cls(src.opname, src.opmap, src.hascompare, src.hasconst,
                     src.hasfree, src.hasjabs, src.hasjrel, src.haslocal,
                     src.hasname, getattr(src, 'hasnargs', ()), no_return,
                     src.HAVE_ARGUMENT)
        for index, opname in enumerate(src.opname):
            if '<' not in opname:
                options = result.get_options(index)
                options = dict((option, True) for option in options)
                result.operations[opname] = OperationType(index, opname, options)
        return result
