# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from entangled.base import AnalysisException
from entangled.operation.definitions import Op, Constant, ArgList

HANDLERS = {}
OPARG = object()


def constant_if_none(value):
    return Constant(value) if value is None else value


def defop(name, pop=0, push=0, read=None, shuffle=None, pass_oparg=False,
          into=None, may_raise=False):
    def OPERATE(f, oparg, frame):
        rest, args = frame.stack.popn(oparg if pop is OPARG else pop)
        if shuffle:
            args = [args[i] for i in shuffle]
        if pass_oparg == "namei":
            args.insert(0, Constant(f.code.codeobj.co_names[oparg]))
        elif pass_oparg:
            args.insert(0, Constant(oparg))
        op = frame.op(name, args)
        if push == 0:
            stack = rest
            mem = op.proj_M # because proj_C is a thing
        elif push == 1:
            stack = rest.cons(op.proj_V)
            mem = op.proj_M
        elif push == OPARG:
            stack = rest.extend(getattr(op, 'proj_%d' % i)
                                for i in xrange(oparg))
            mem = op.proj_M
        else:
            raise AnalysisException("Need more push?")
        f.custom_set(None, mem, op.proj_C, stack=stack)
        if may_raise:
            f.may_raise(frame, op)
    if into:
        into[name] = OPERATE
    else:
        HANDLERS[name] = OPERATE

defop('STORE_SUBSCR', 3, shuffle=(1, 2, 0), may_raise=True)
defop('DELETE_SUBSCR', 3, shuffle=(1, 2, 0), may_raise=True)
defop('PRINT_EXPR', 1)
defop('PRINT_ITEM_TO', 2) # TOS is stream which ends up second
defop('PRINT_ITEM', 1)
defop('PRINT_NEWLINE_TO', 1)
defop('PRINT_NEWLINE', 0)
defop('BUILD_CLASS', 3, 1, shuffle=(2, 1, 0), may_raise=True)
defop('UNPACK_SEQUENCE', 1, OPARG, may_raise=True)
defop('STORE_ATTR', 2, shuffle=(1, 0), pass_oparg="namei")
defop('DELETE_ATTR', 1, pass_oparg="namei", may_raise=True)
defop('LOAD_ATTR', 1, 1, pass_oparg="namei", may_raise=True)
defop('STORE_GLOBAL', 2, pass_oparg="namei")
defop('DELETE_GLOBAL', pass_oparg="namei", may_raise=True)
defop('LOAD_GLOBAL', 0, 1, pass_oparg="namei", may_raise=True)
defop('BUILD_TUPLE', OPARG, 1)
defop('BUILD_LIST', OPARG, 1)
defop('COMPARE_OP', 2, 1, pass_oparg=True, may_raise=True)
defop('IMPORT_NAME', 2, 1, pass_oparg=True, may_raise=True)
defop('IMPORT_FROM', 1, 1, pass_oparg=True, may_raise=True)
defop('GET_ITER', 1, 1, may_raise=True)
defop('GET_AITER', 1, 1, may_raise=True)
defop('GET_AWAITABLE', 1, 1, may_raise=True)


def handler(f):
    assert f.__name__ not in HANDLERS, f.__name__
    HANDLERS[f.__name__] = f
    return f


def handle(name):
    def decorate(f):
        HANDLERS[name] = f
        return f
    return decorate


def defnoop(f):
    @functools.wraps(f)
    def wrapper(op_state, oparg, frame):
        op_state.no_jump(trans=f)
    HANDLERS[f.__name__] = wrapper
    return wrapper


@handler
def LOAD_CONST(f, oparg, frame):
    value = f.code.codeobj.co_consts[oparg]
    op = frame.op('LOAD_CONST', [Constant(oparg), Constant(value)])
    f.set_next(op, stack=frame.stack.cons(op))
    #no_except=True)

@handler
def LOAD_FAST(f, oparg, frame):
    # XXX: check for unset values
    value = frame.locals_[oparg]
    name = Constant(f.code.codeobj.co_varnames[oparg])
    op = frame.op('LOAD_FAST', [name, value])
    f.set_next(op, stack=frame.stack.cons(op.proj_V))
    #no_except=True)
    f.may_raise(frame, op)

@handler
def STORE_FAST(f, oparg, frame):
    rest, value = frame.stack.decons()
    name = Constant(f.code.codeobj.co_varnames[oparg])
    op = frame.op('STORE_FAST', [name, value])
    locals_ = frame.local[:]
    locals_[oparg] = op.proj_V
    f.set_next(op, stack=rest, local=locals_)

@handler
def DELETE_FAST(f, oparg, frame):
    # XXX: check for unset values
    name = Constant(f.code.codeobj.co_varnames[oparg])
    op = frame.op('DELETE_FAST', [name])
    locals_ = frame.locals_[:]
    locals_[oparg] = op
    f.set_next(op, locals_=locals_)
    #f.may_raise(frame, op) TODO: simplify
    exc_info = [op.proj_EX_TY, op.proj_EX_V, op.proj_EX_TB]
    f.jump_to_handler(frame, op, f.WHY_EXCEPTION, exc_info)

@defnoop
def POP_TOP(f, unused, frame):
    rest, _ = frame.stack.decons()
    return {'stack': rest}

@defnoop
def ROT_TWO(f, unused, frame):
    rest, [y, z] = frame.stack.popn(2)
    return {'stack': rest.extend([z, y])}

@defnoop
def ROT_THREE(f, unused, frame):
    rest, [x, y, z] = frame.stack.popn(3)
    return {'stack': rest.extend([z, x, y])}

@defnoop
def ROT_FOUR(f, unused, frame):
    rest, [w, x, y, z] = frame.stack.popn(4)
    return {'stack': rest.extend([z, w, x, y])}

@defnoop
def DUP_TOP(f, unused, frame):
    top = frame.stack.get(-1)
    return {'stack': stack=frame.stack.cons(top)}

@defnoop
def DUP_TOP_TWO(f, unused, frame):
    _, [y, z] = frame.stack.popn(-2)
    return {'stack': frame.stack.extend([y, z])}

@defnoop
def DUP_TOPX(f, num, frame):
    _, last_items = frame.stack.popn(num)
    return {'stack': frame.stack.extend(last_items)}


def get_slice(stack, n):
    has_start = n & 1
    has_stop = (n >> 1) & 1
    start = stop = Constant(None)
    if has_stop:
        stack, stop = stack.decons()
    if has_start:
        stack, start = stack.decons()
    return stack, start, stop


def slicer(n):
    def SLICE(f, unused, frame):
        stack = frame.stack
        stack, start, stop = get_slice(stack, n)
        stack, base = stack.decons()
        op = frame.op("SLICE", [base, start, stop])
        f.set_next(op, stack=stack.cons(op.proj_V))
        f.may_raise(frame, op)
    return SLICE

def store_slice(n):
    def STORE_SLICE(f, unused, frame):
        stack, value = frame.stack.pop()
        stack, start, stop = get_slice(stack, n)
        stack, base = stack.decons()
        op = frame.op("STORE_SLICE", [base, start, stop, value])
        f.set_next(op, stack=stack)
        f.may_raise(frame, op)
    return STORE_SLICE

def delete_slice(n):
    def DELETE_SLICE(f, unused, frame):
        stack = frame.stack
        stack, start, stop = get_slice(stack, n)
        stack, base = stack.decons()
        op = frame.op("DELETE_SLICE", [base, start, stop])
        f.set_next(op, stack=stack)
        f.may_raise(frame, op)
    return DELETE_SLICE

HANDLERS['SLICE+0'] = slicer(0)
HANDLERS['SLICE+1'] = slicer(1)
HANDLERS['SLICE+2'] = slicer(2)
HANDLERS['SLICE+3'] = slicer(3)
HANDLERS['STORE_SLICE+0'] = store_slice(0)
HANDLERS['STORE_SLICE+1'] = store_slice(1)
HANDLERS['STORE_SLICE+2'] = store_slice(2)
HANDLERS['STORE_SLICE+3'] = store_slice(3)
HANDLERS['DELETE_SLICE+0'] = delete_slice(0)
HANDLERS['DELETE_SLICE+1'] = delete_slice(1)
HANDLERS['DELETE_SLICE+2'] = delete_slice(2)
HANDLERS['DELETE_SLICE+3'] = delete_slice(3)

@handler
def LIST_APPEND(f, to_list, frame):
    xs = frame.stack.get(-to_list)
    rest, x = frame.stack.decons()
    op = frame.op('LIST_APPEND', [xs, x])
    f.set_next(op, stack=rest)
    # while we must permit GC to run here (which means running arbitrary
    # finalisers) it would be nice to remove
    #f.may_raise(frame, op)

@handler
def SET_ADD(f, to_list, frame):
    xs = frame.stack.get(-to_list)
    rest, x = frame.stack.decons()
    op = frame.op('SET_ADD', [xs, x])
    f.set_next(op, stack=rest)
    f.may_raise(frame, op)

@handler
def RAISE_VARARGS(f, nargs, frame):
    rest, args = frame.stack.popn(nargs)
    exc_info = args + [frame.op('NONE', [])] * (3 - nargs)
    f.jump_to_handler(frame, frame.op('RAISE_VARARGS', args), f.WHY_EXCEPTION,
                      exc_info=exc_info, stack=rest)

@handler
def LOAD_LOCALS(f, unused, frame):
    op = frame.op('LOAD_LOCALS', frame.locals_)
    f.set_next(op, stack=frame.stack.cons(op))

@handler
def RETURN_VALUE(f, unused, frame):
    rest, value = frame.stack.decons()
    f.jump_to_handler(frame, frame.op('RETURN_VALUE', [value]), f.WHY_RETURN)

@handler
def YIELD_FROM(f, unused, frame):
    rest, value = frame.stack.decons()
    op = frame.op('YIELD_FROM', [value])
    f.yield_to_handler(op, value, frame)
    f.set_next(op, stack=rest.cons(op.proj_V))
    f.may_raise(frame, op)

# XXX todo: I think python 2.5 didn't push a value to the stack on
# yield.  See what we can do about this.
@handler
def YIELD_VALUE(f, unused, frame):
    rest, value = frame.stack.decons()
    op = frame.op('YIELD_VALUE', [value])
    f.yield_to_handler(op, value, frame)
    f.set_next(op, stack=rest.cons(op.proj_V))
    f.may_raise(frame, op)

@handler
def EXEC_STMT(f, unused, frame):
    rest, [u, v, w] = frame.stack.popn(3)
    op = frame.op('EXEC_STMT', [u, v, w, frame.locals_])
    f.set_next(op, stack=rest)
    f.may_raise(frame, op)

# XXX todo: _NAME opcodes are wierd, decide what to do
# probably check statically what paths arrive here
@handler
def STORE_NAME(f, namei, frame):
    rest, value = frame.stack.decons()
    op = frame.op('STORE_NAME', [Constant(namei), value])
    f.set_next(op, rest)
    f.may_raise(frame, op)

@handler
def DELETE_NAME(f, namei, frame):
    op = frame.op('DELETE_NAME', [Constant(namei)])
    f.set_next(op)
    f.may_raise(frame, op)

@handler
def LOAD_NAME(f, namei, frame):
    op = frame.op('LOAD_NAME', [Constant(namei)])
    f.set_next(op, stack=frame.stack + [op.proj_V])
    f.may_raise(frame, op)

@handler
def LOAD_CLOSURE(f, freevari, frame):
    var = frame.freevars[freevari]
    op = frame.op('LOAD_CLOSURE', [Constant(freevari), var])
    f.set_next(op, stack=frame.stack + [op.proj_V])
    f.may_raise(frame, op)

@handler
def LOAD_DEREF(f, freevari, frame):
    var = frame.freevars[freevari]
    op = frame.op('LOAD_DEREF', [Constant(freevari), var])
    f.set_next(op, stack=frame.stack + [op.proj_V])
    f.may_raise(frame, op)

@handler
def STORE_DEREF(f, freevari, frame):
    rest, val = frame.stack.decons()
    freevars = frame.freevars[:]
    var = freevars[freevari]
    op = frame.op('STORE_DEREF', [var, val])
    freevars[freevari] = op # ?
    f.set_next(op, freevars=freevars)

############################################################
# Building Literals

@handler
def BUILD_MAP(f, oparg, frame):
    if f.config.advanced_build_map:
        rest, items = frame.stack.popn(oparg * 2)
        op = frame.op('BUILD_MAP.A', items)
        f.may_raise(frame, op)
    else:
        rest = frame.stack
        op = frame.op('BUILD_MAP', [])
    f.set_next(op, stack=rest.cons(op.proj_V)])

# XXX do these two really use the opcode differently?
@handler
def STORE_MAP(f, unused, frame):
    rest, [value, key] = frame.stack.popn(2)
    op = frame.op('STORE_MAP', [rest.get(-1), key, value])
    #frame.set_from(f, frame, op, stack=rest)
    f.set_next(op, stack=rest)
    f.may_raise(frame, op)

@handler
def MAP_ADD(f, map_offset, frame):
    stack, [value, key] = frame.stack.popn(2)
    op = frame.op('MAP_ADD', [frame.stack.get(-map_offset), key, value])
    #frame.set_from(f, frame, op, stack=stack)
    f.set_next(op, stack=rest)
    f.may_raise(frame, op)

@handler
def IMPORT_STAR(f, unused, frame):
    rest, mod = frame.stack.decons()
    op = frame.op('IMPORT_STAR', [mod])
    locals_ = [frame.op('entangled.python.import_star_maybe',
                    [op, Constant(namei), prev])
               for namei, prev in enumerate(frame.locals_)]
    f.set_next(op, locals_=locals_) # XXX update stack?
    f.may_raise(frame, op)

############################################################
# blocks!

def block_setup_handler(name):
    def SETUP_BLOCK(f, oparg, frame):
        target = f.code.by_offset[frame.jump_target]
        # XXX should store oparg here also
        op = frame.op(name, [Constant(target), Constant(len(frame.stack))])
        # note that there is no blockstack at this stage of complation.
        f.set_next(op)
        f.add_scope(f.source_op.proj_stack_size, op.proj_stack_size)
    HANDLERS[name] = SETUP_BLOCK
    return SETUP_BLOCK

block_setup_handler('SETUP_LOOP')
block_setup_handler('SETUP_EXCEPT')
block_setup_handler('SETUP_FINALLY')

@handler
def SETUP_WITH(f, oparg, frame):
    stack, context_manager = frame.stack.decons()
    op = frame.op('SETUP_WITH',
                  [Constant(oparg), Constant(len(stack) + 2), context_manager])
    # note that there is no blockstack at this stage of compilation.
    f.set_next(op, stack=stack.cons(op.proj_exit).cons(op.proj_enter))

@handler
def WITH_CLEANUP_START(f, oparg, frame):
    rest, [exitfunc, unroller] = f.stack.popn(2)
    stack = rest.cons(unroller)
    offset = f.get_offset()
    is_exc = f.op('is_sapplicationexception', [unroller])

    NONE = Constant(None)

    exc_case = f.graph_state.target_graph.block()
    exc_case.add_predecessor(is_exc.proj_T)
    exc = exc_case.op('sapplicationexception_destructure', offset,
                      [is_exc.proj_M], [unroller])
    args = ArgList([exc.proj_EX_V, exc.proj_EX_TY, exc.proj_EX_TB])
    call_t = exc_case.op('CALL_FUNCTION', offset, [is_exc.proj_M],
                         [NONE, args, ArgList([]), NONE, NONE])
    jump_t = exc_case.op('entangled.true', [], []).proj_C

    norm_case = f.graph_state.target_graph.block()
    norm_case.add_predecessor(is_exc.proj_F)
    args = ArgList([NONE, NONE, NONE])
    call_f = norm_case.op('CALL_FUNCTION', offset, [is_exc.proj_M],
                          [NONE, args, ArgList([]), NONE, NONE])
    jump_f = exc_case.op('entangled.true', [], []).proj_C

    result = f.graph_state.target_graph.block()
    result.add_predecessor(jump_t)
    result.add_predecessor(jump_f)
    mem = result.op('entangled.phi.mem', offset,
                    [call_t.proj_M, call_f.proj_M], [])
    value = result.op('entangled.phi.result', offset,
                      [call_t.proj_V, call_f.proj_V], [])
    jump_r = result.op('entangled.true', [], []).proj_C
    f.custom_set(None, mem, jump_r, stack=rest.extend([unroller, value]))


@handle('WITH_CLEANUP')
@handler
def WITH_CLEANUP_FINISH(f, oparg, frame):
    rest, [unroller, suppress] = f.stack.popn(2)
    is_suppressed = f.op('JUMP', [suppress])
    
    case = f.graph_state.target_graph.block()
    case.add_predecessor(is_suppressed.proj_T)
    case.add_predecessor(is_suppressed.proj_F)
    result = case.op('entangled.phi.result', f.get_offset(), [],
                     [Constant(None), unroller])
    end = case.op('entangled.true', [], []).proj_C

    f.custom_set(None, is_suppressed.proj_M, end, stack=rest.cons(result))

@handler
def WITH_CLEANUP(f, oparg, frame, body):
    # yech.  much like the finally situation, we're not sure how big
    # the stack is.
    pass

@defnoop
def POP_EXCEPT(f, unused, frame):
    stack, _ = frame.stack.popn(3)
    # note that there is no blockstack at this stage of compilation.
    return {'stack': stack}
    # may_raise?

@defnoop
def POP_BLOCK(f, unused, frame):
    # note that there is no blockstack at this stage of compilation.
    # do we need to restore an exception?
    return {}

@defnoop
def BEGIN_FINALLY(f, unused, frame):
    return {'stack': frame.stack.cons(frame.op('NONE', []))}

@handler
def POP_FINALLY(f, is_return, frame):
    """End a finally block.

    Statically indicates if there is a return value on the stack,
    which means the return value can appear above the WHY.

    Availability: Python 3.6+

    In CPython, POP_FINALLY also pops EXCEPT_HANDLER, aka
    RestoreSysExc.

    """
    pop_finally_implementation(f, is_return, frame, 'POP_FINALLY', True)

@handler
def END_FINALLY(f, unused, frame):
    """End a finally block.

    The first item on the stack, if it is not an exception, is a WHY
    value indicating if there is a return value present.

    In recent CPython, this uses _PyErr_Restore on the top exception.

    """
    pop_finally_implementation(f, None, frame, 'END_FINALLY', False)


def pop_finally_implementation(f, is_return, frame, name, popsix):
    """Common implementation of [pop|end]_finally opcodes

    We follow the stack discipline of pypy: there is one value on the
    stack we need to pop; called the unroller.

    Our implementation is fairly abstract: we produce an operation for
    dispatching on the unroller, and bind its proj with the
    appropriate jump_to_handler or set_next.

    The implementation of the abstract operation is expected to
    perform any SysExcRestorer functionality required.

    IWBNI we knew statically what reasons were valid at this opcode.

    """
    unroller, rest = frame.stack.decons()
    op = f.op(name, [frame.mem], [unroller])
    f.set_next(op, stack=rest)
    # we build three outgoing edges; we can cull them in the next pass
    # when we've determined the complete list of unrollers.
    f.custom_set('raise', op.proj_M, op.proj_raise, stack=rest)
    f.custom_set('break', op.proj_M, op.proj_break, stack=rest)
    f.custom_set('return', op.proj_M, op.proj_return, stack=rest)


def old_pop_finally_implementation(f, is_return, frame, name, popsix):
    if is_return not in (0, 1, None):
        raise AnalysisException("Bad POP_FINALLY oparg")
    stack = frame.stack
    return_value = None
    if is_return:
        stack, return_value = stack.decons()
    # we set up the blocks so that the indicator is known statically
    indicator = stack.get(-1)
    # XXX: once we get stack splitting working, we'll need to change
    # this `if`, and the ones below, into a different beast.
    if is_exception_type(indicator):
        if popsix:
            # drop the top 3 values in the new opcode
            stack, _ = stack.popn(3)
        stack, exc_info = stack.popn(3)
        f.jump_to_handler(frame, frame.op(name, exc_info[:]),
                          f.WHY_EXCEPTION, exc_info[::-1],
                          stack=stack,
                          blockstack=frame.blocks[:-1],
                          retval=return_value)
        return
    else:
        # non-exceptional control flow into block
        stack, indicator = stack.decons()
        if is_why(indicator) and is_return is None:
            # in older versions of python, we consult the indicator as
            # to the presence of a return value.
            if constant_lookup(indicator).issubset(
                    set([f.WHY_RETURN, f.WHY_CONTINUE])):
                stack, return_value = stack.decons()
        op = frame.op(name, [indicator, return_value])
        if is_none(indicator) or is_why(indicator):
            f.set_next(frame, stack, op, retval=return_value,
                       blockstack=frame.blocks[:-1])
            return
        raise AnalysisException(repr(indicator))
    raise AnalysisException(repr(indicator))


@handler
def CALL_FINALLY(f, oparg, frame):
    f.set_alt(frame, stack=frame.stack.cons(Constant(oparg)))

############################################################
# Calling functions & methods

@handler
def LOOKUP_METHOD(f, namei, frame):
    """pypy-specific optimisation

    Avoids allocating a bound method in the typical case
    """
    stack, obj = frame.stack.decons()
    op = frame.op('LOOKUP_METHOD', [Constant(namei), obj])
    f.set_next(op, stack=stack.extend([op.proj_V, obj]))
    # may_raise if lookup raises
    f.may_raise(frame, op) 

@handler
def CALL_METHOD(f, oparg, frame):
    stack, w_self = frame.stack.decons()
    call_function(stack, f, oparg, frame, w_self=w_self)

@handler
def CALL_FUNCTION(f, oparg, frame):
    call_function(frame.stack, f, oparg, frame)

@handler
def CALL_FUNCTION_VAR(f, oparg, frame):
    stack, varargs = frame.stack.decons()
    call_function(stack, f, oparg, frame, varargs)

@handler
def CALL_FUNCTION_KW(f, oparg, frame):
    stack, varkw = frame.stack.decons()
    call_function(stack, f, oparg, frame, varkw=varkw)

@handler
def CALL_FUNCTION_VAR_KW(f, oparg, frame):
    stack, [varargs, varkw] = frame.stack.popn(2)
    call_function(stack, f, oparg, frame, varargs, varkw)

def call_function(stack, f, oparg, frame, varargs=None, varkw=None,
                  w_self=None):
    na = oparg & 0xff
    nk = (oparg >> 8) & 0xff
    to_consume = na + 2 * nk

    stack, kw_items = stack.popn(nk * 2)
    kw_list = list(zip(*[iter(kw_items) * 2]))
    stack, pos_args = stack.popn(na)
    stack, target = stack.decons()

    op = frame.op('CALL_FUNCTION',
                  [target, constant_if_none(w_self), ArgList(pos_args),
                   ArgList(kw_list), constant_if_none(varargs),
                   constant_if_none(varkw)])
    f.set_next(op, stack=stack.cons(op.proj_V))
    f.may_raise(frame, op)

@handler
def MAKE_FUNCTION(f, oparg, frame):
    stack, contents = frame.stack.popn(1 + oparg)
    code = contents.pop()
    defaults = None
    if oparg:
        defaults = frame.op('BUILD_TUPLE', contents)
    op = frame.op('MAKE_FUNCTION', [Constant(oparg), code, defaults])
    f.set_next(op, stack=stack.cons(op.proj_V))

@handler
def MAKE_CLOSURE(f, oparg, frame):
    stack, contents = frame.stack.popn(2 + oparg)
    code = contents.pop()
    closure = contents.pop()
    defaults = None
    if oparg:
        defaults = frame.op('BUILD_TUPLE', contents)
    op = frame.op('MAKE_CLOSURE', [Constant(oparg), code, closure, defaults])
    f.set_next(op, stack=stack.cons(op.proj_V))

@handler
def BUILD_SLICE(f, oparg, frame):
    if oparg == 3:
        stack, [u, v, w] = frame.stack.popn(3)
    else:
        w = None
        stack, [u, v] = frame.stack.popn(2)
    op = frame.op('BUILD_SLICE', [u, v, w])
    f.set_next(op, stack=stack.cons(op.proj_V))
    # may_raise?

@defnoop
def EXTENDED_ARG(f, oparg, frame):
    return {}

def add_standard_opcodes(handlers, config):
    for opname in config.opname:
        if opname.startswith('INPLACE_'):
            defop(opname, 2, 1, into=handlers)
        elif opname.startswith('BINARY_'):
            defop(opname, 2, 1, into=handlers)
        elif opname.startswith('UNARY_'):
            defop(opname, 1, 1, into=handlers)

# def handle_inplace_op(handlers, opname, inplace_name, left_name, right_name):
#     def INPLACE(f, oparg, frame):
#         rest, [item, value] = f.stack.popn(2)
#         offset = f.get_offset()
#         sm_inplace = f.op('lookup_method', [item, Constant(inplace_name)])
#         sm_left = f.op('lookup_method', [item, Constant(left_name)])
#         sm_right = f.op('lookup_method', [item, Constant(right_name)])

#         has_inplace = f.op('is_none', [sm_inplace])

#         inplace = f.graph_state.target_graph.block()
#         inplace.add_predecessor(has_inplace.proj_F)
#         inplace_call = inplace.op('call_sm', offset, [sm_inplace.proj_M],
#                                   [sm_inplace, item, ArgList([value])])

#         inplace_check = f.graph_state.target_graph.block()
#         inplace.add_predecessor(inplace_call.proj_C)
#         inplace_implemented = inplace_check.op('is_notimplemented', offset,
#                                                [inplace_call.proj_M],
#                                                [inplace_call.proj_V])

#         left = f.graph_state.target_block.block()
#         left.add_predecessor(has_inplace.proj_T)
#         left.add_predecessor(inplace_implemented.proj_T)
#         mem = left.op('entangled.phi.mem', offset,
#                       [sm_left.proj_M, inplace_call.proj_M], [])
#         left_call = left.op('call_sm', offset, [mem.proj_M],
#                             [sm_left, item, ArgList([value])])

#         left_check = f.graph_state.target_graph.block()
#         left.add_predecessor(left_call.proj_C)
#         left_implemented = left_check.op('is_notimplemented', offset,
#                                          [left_call.proj_M],
#                                          [left_call.proj_V])

#         # return
#         result = f.graph_state.target_graph.block()
#         result.add_predecessor(inplace_implemented.proj_F)
#         result.add_predecessor(left_implemented.proj_F)
#         result.add_predecessor(right_implemented.proj_F)
#         result_value = result.op('entangled.phi.result', offset,
#                                  [], [inplace_call.proj_V,
#                                       left_call.proj_V,
#                                       right_call.proj_V])


#         # reraise - or should we raise immediately?
#         reraise = f.graph_state.target_graph.block()
#         reraise.add_predecessor(inplace_call.proj_raise_C)
#         reraise.add_predecessor(left_call.proj_raise_C)
#         reraise.add_predecessor(right_call.proj_raise_C)
#     return INPLACE

############################################################
# Control flow

@defnoop
def NOP(f, oparg, frame):
    return {}

@handler
def JUMP_ABSOLUTE(f, oparg, frame):
    f.set_alt(frame)

@handler
def JUMP_FORWARD(f, oparg, frame):
    f.set_alt(frame)

@handler
def JUMP_IF_NOT_DEBUG(f, oparg, frame):
    """A pypy-specific opcode
    """
    op = frame.op('JUMP_IF_NOT_DEBUG', [])
    f.set_next(frame, op.proj_M, op.proj_F)
    f.set_alt(frame, op.proj_M, op.proj_T)

@handler
def POP_JUMP_IF_FALSE(f, oparg, frame):
    stack, value = frame.stack.decons()
    op = frame.op('JUMP', [value])
    f.set_next(op, 'T', stack)
    f.set_alt(frame, op.proj_M, op.proj_F, stack)
    f.may_raise(frame, op)

@handler
def POP_JUMP_IF_TRUE(f, oparg, frame):
    stack, value = frame.stack.decons()
    op = frame.op('JUMP', [value])
    f.set_next(op, 'F', stack)
    f.set_alt(frame, op.proj_M, op.proj_T, stack)
    f.may_raise(frame, op)

@handler
def JUMP_IF_FALSE_OR_POP(f, oparg, frame):
    rest, check = frame.stack.decons()
    op = frame.op('JUMP', check)
    f.set_next(op, 'T', stack=rest)
    f.set_alt(frame, op.proj_M, op.proj_F)
    f.may_raise(frame, op)

@handler
def JUMP_IF_TRUE_OR_POP(f, oparg, frame):
    rest, check = frame.stack.decons()
    op = frame.op('JUMP', check)
    f.set_next(op, 'F', stack=rest)
    f.set_alt(frame, op.proj_M, op.proj_T)
    f.may_raise(frame, op)

@handler
def GET_ANEXT(f, oparg, frame):
    rest, iterator = frame.stack.decons()
    op = frame.op('GET_ANEXT', [iterator])
    f.set_next(op, stack=frame.stack.cons(op.proj_V))
    f.may_raise(frame, op)

@handler
def FOR_ITER(f, oparg, frame):
    rest, iterator = frame.stack.decons()
    op = frame.op('FOR_ITER', [Constant(oparg), iterator])
    f.set_next(op, stack=frame.stack.cons(op.proj_V))
    f.set_alt(frame, op.proj_M, stack=rest)
    f.may_raise(frame, op)
#    assert f.frames_[frame.offset + 1, frame.shadow].

@handler
def BREAK_LOOP(f, unused, frame):
    f.jump_to_handler(frame, frame.op('BREAK_LOOP', []), f.WHY_BREAK)

@handler
def CONTINUE_LOOP(f, oparg, frame):
    f.jump_to_handler(frame, frame.op('CONTINUE_LOOP', [oparg]),
                      f.WHY_CONTINUE)

# On handling the various block unrollers:
#
# the SContinueLoop unroller follows the `alt` prefix of the CONTINUE
# op.
#
# the SReturnValue unroller follows the `return` prefix of the
# RETURN_VALUE op.
#
# the SBreakLoop unroller follows the `break` prefix of the BREAK_LOOP op.
#
# the SApplicationException unroller follows the `exc` prefix of any
# op that may_raise.
#
# the YIELD_VALUE and YIELD_FROM operations do not unroll; instead
# they are added to a graph-specific yields list.
#
# The unrollers are stored at frame.unroller.  Handlers typically push
# this value onto the stack for cleaning up later.  Speaking of, it is
# the responsibility of the handler to set up the stack appropriately.

@handle('landing.loop')
@handle('landing.finally')
@handle('landing.return')
def LANDING(f, frame):
    op = f.op('nop', [frame.mem], [])
    f.set_next(op, stack=frame.stack.cons(frame.unroller))

@handle('landing.except')
def LANDING_EXCEPT(f, frame):
    unroller = frame.unroller
    op = f.op('sapplicationexception_destructure', [frame.mem], [unroller])
    f.set_next(op, stack=frame.stack.extend(
        [frame.unroller, op.proj_EX_TB, op.proj_EX_TY, op.proj_EX_V]
    ))
