# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import attr
import dis

# Opcodes that have no automatic control flow to the next operation.
STANDARD_NO_RETURN = (
    'BREAK_LOOP', 'CONTINUE_LOOP', 'JUMP_ABSOLUTE', 'JUMP_FORWARD',
    'RAISE_VARARGS', 'RETURN_VALUE', 'STOP_CODE'
)

class Config(object):
    def __init__(self, opname, opmap, hascompare=(), hasconst=(),
                 hasfree=(), hasjabs=(), hasjrel=(), haslocal=(),
                 hasname=(), hasnargs=(), no_return=(), have_argument=None):
        self.opname = opname
        self.opmap = opmap
        self.hascompare = hascompare
        self.hasconst = hasconst
        self.hasfree = hasfree
        self.hasjabs = hasjabs
        self.hasjrel = hasjrel
        self.haslocal = haslocal
        self.hasname = hasname
        self.hasnargs = hasnargs
        self.no_return = no_return
        self.HAVE_ARGUMENT = have_argument

    def has_argument(self, opno):
        return opno > self.HAVE_ARGUMENT

    @classmethod
    def from_native(cls, src=dis):
        no_return = [src.opmap[name] for name in STANDARD_NO_RETURN
                     if name in src.opmap]
        return cls(src.opname, src.opmap, src.hascompare, src.hasconst,
                   src.hasfree, src.hasjabs, src.hasjrel, src.haslocal,
                   src.hasname, getattr(src, 'hasnargs', ()), no_return,
                   src.HAVE_ARGUMENT)


@attr.s
class NonlocalType(object):
    name = attr.ib()
    handlers = attr.ib()


# possible cause of exit -> handlers that might catch it
RETURN = NonlocalType('return', ['return', 'finally'])
CONTINUE = NonlocalType('continue', ['loop', 'finally'])
BREAK = NonlocalType('break', ['loop', 'finally'])
EXCEPT = NonlocalType('except', ['except', 'finally'])


class BytecodeOperation(object):
    """Represent a bytecode operation.

    Bytecode operations are fairly context-independent; they have

    :param: config    description of bytecode format
    :param: offset_bc the bytecode offset of the end of the instruction
    :param: offset    the index of the instruction in the stream
    :param: op        the numeric value of the operation
    :param: oparg     the argument given (or None)
    """
    def __init__(self, config, offset_start, offset_end, offset, op, oparg):
        self.config = config
        self.offset_start = self.offset_bc = offset_start
        self.offset_end = offset_end
        self.offset = offset
        self.op = op
        self.oparg = oparg

    @property
    def opname(self):
        return self.config.opname[self.op]

    @property
    def hasjrel(self):
        return self.op in self.config.hasjrel

    @property
    def hasjabs(self):
        return self.op in self.config.hasjabs

    @property
    def jump_target(self):
        if self.hasjrel:
            return self.offset_end + self.oparg
        if self.hasjabs:
            return self.oparg
        return None

    @property
    def no_return(self):
        return self.opname in self.config.no_return

    @property
    def nonlocal_types(self):
        if self.opname == 'CONTINUE':
            return [CONTINUE]
        if self.opname == 'BREAK':
            return [BREAK]
        if self.opname == 'RETURN':
            return [RETURN]
        if self.opname in REMOVES_BLOCK:
            result = [CONTINUE, BREAK, RETURN, EXCEPT]
        return [EXCEPT]

    def __repr__(self):
        return "%s<%s>(%s)" % (self.opname, self.offset, self.oparg)

    def __str__(self):
        return self.opname


class Code(object):
    def __init__(self, codeobj, config, operations, labels, by_offset):
        self.codeobj = codeobj
        self.config = config
        self.operations = operations
        self.labels = labels
        self.by_offset = by_offset

    @classmethod
    def from_bytecode(cls, codeobj, config=None):
        if config is None:
            config = Config.from_native()
        operations, labels, by_offset = to_operations(config, codeobj.co_code)
        return cls(codeobj, config, operations, labels, by_offset)

    def __iter__(self):
        return iter(self.operations)

    def __len__(self):
        return len(self.operations)

    def __getitem__(self, index):
        # op = self.operations[index]
        # targets = []
        # if not op.no_return:
        #     targets.append(index + 1)
        # if op.hasjrel:
        #     targets.append(self.by_offset[op.jump_target])
        # return op, self.labels.get(op.offset, ()), targets
        return self.operations[index]

    def iteritems(self):
        for index, op in enumerate(self.operations):
            targets = []
            if not op.no_return:
                targets.append(index + 1)
            if op.hasjrel:
                targets.append(self.by_offset[op.jump_target])
            yield op, self.labels.get(op.offset, ()), targets


def to_operations(config, code):
    i = 0
    operations = []
    labels = {}
    by_offset = {}
    extended_arg = 0
    while i < len(code):
        op = ord(code[i])
        current = i
        offset = len(operations)
        i += 1
        if config.has_argument(op):
            oparg = ord(code[i]) + ord(code[i + 1]) * 256 + extended_arg
            i += 2
            if op == config.opmap['EXTENDED_ARG']:
                extended_arg = oparg * 65536
            else:
                op = BytecodeOperation(config, current, i, offset, op, oparg)
                operations.append(op)
                by_offset[current] = offset
                extended_arg = 0
                if op.hasjrel:
                    labels.setdefault(oparg + i, []).append(current)
                elif op.hasjabs:
                    labels.setdefault(oparg, []).append(current)
        else:
            if extended_arg:
                raise AnalysisException("Extended arg for argless operation")
            by_offset[current] = offset
            operations.append(BytecodeOperation(config, current, i, offset,
                                                op, None))
    return operations, labels, by_offset


REMOVES_BLOCK = set(
    "POP_EXCEPT POP_BLOCK POP_FINALLY END_FINALLY END_ASYNC_FOR "
    "WITH_CLEANUP_FINISH".split())
ADDS_BLOCK = {
    "SETUP_FINALLY" : "SETUP_FINALLY",
    "SETUP_ASYNC_WITH" : "SETUP_FINALLY",
    "SETUP_WITH" : "SETUP_FINALLY",
    "SETUP_LOOP" : "SETUP_LOOP",
    "SETUP_EXCEPT" : "SETUP_EXCEPT"
#    "FOR_ITER" : "", should pop block?
}


def _set_target_blockstack(blockstacks, target, blockstack):
    follow_stack = blockstacks[target]
    if follow_stack is None:
        blockstacks[target] = blockstack
    else:
        if follow_stack != blockstack:
            raise AnalysisError('Inconsistent blockstack?')


def discover_blockstack(code, labels):
    blockstacks = [None] * len(code.operations)
    blockstacks[0] = blockstack
    for index, (op, preds, targets) in enumerate(code.iteritems()):
        blockstack = blockstacks[index]

        if op.opname in ADDS_BLOCK:
            # for instructions that add a block, the first target is
            # usually the next instruction, and the other target is
            # the handler which operates without the block present.
            #
            # should check that this is the case.
            assert len(targets) == 2
            assert target[0].offset <= (op.offset + 5)
            new_item = (ADDS_BLOCK[op.opname], target[1].offset)
            _set_target_blockstack(blockstacks, target[0],
                                   blockstack + [new_item])
            _set_target_blockstack(blockstacks, target[1], blockstack)
            continue
        elif op.opname in REMOVES_BLOCK:
            blockstack = blockstack[:-1]

        for target in targets:
            _set_target_blockstack(blockstacks, target, blockstack)
    return blockstacks
            
        

