# Copyright (c) 2019 William ML Leslie

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Convert from trivial-ssa to a full sea-of-nodes SSA.

This process proceeds by iterating over the nodes and applying a
handler to them.  What does the handler get?

f - the analysis context.  in the original implementation this was
global to the analysis, but handlers3 uses an f that has some more
specific details.

oparg - the argument passed to the operation, which is sometimes used
to determine the stack effect.

frame - the abstract memory state at the point of the operation.

Must look at the bop definition to see how we can pull any remaining
information out.

Ops may have the following proj:

alt_[CVM] - alternate exit (eg, conditional jumps and loops)
exit_[CVM] - exit block (?)

operations that pop a block seem to have one or more of the following
sets of projections.  My intuition is that raise_, return_, and
control_ would be better.

except_[CVM] - exit by raising
finally_[CVM] - exit by raising, breaking, or returning
loop_[CVM] - exit to a loop handler by breaking or continuing

Must remember to specially connect the raise and return blocks of the new graph and the old one.
"""

from entangled.base import AnalysisException
from entangled.generic.analyserx import FrameState
from entangled.generic.analyser3 import BaseAnalyser
from entangled.generic.analyser3 import BlockState as BlockState3
from entangled.generic.analyser3 import GraphState as GraphState3
from entangled.generic.analyser3 import OpState as OpState3
#from entangled.generic.stack import Stack

# basic layout:
# Analyser:  keeps track of work
# GraphState:  keeps track of a source graph / blocks
# BlockState:  keeps track of a source block / operations (scope)
# OpState:  keeps everything relevant for analysing one operation
# FrameState:  keeps track of the trivial-ssa frame content.
# 
# The BlockState maintains both a scope for FrameState use and one for
# merging.


class DumbStack(object):
    """Implement the stack api, using a list to back it.
    """
    def __init__(self, items):
        self.items = items

    def decons(self): # -> top, rest
        return self.items[-1], DumbStack(self.items[:-1])

    def __bool__(self):
        return bool(self.items)

    def popn(self, n): # -> rest, [obs]
        return self.items[-n:], DumbStack(self.items[:-n])

    def trunc(self, n):
        return DumbStack(self.items[:n])

    def cons(self, value):
        return DumbStack(self.items + [value])

    def extend(self, values):
        return DumbStack(self.items + list(values))

    def update(self, block, stack):
        # ensure that depth is consistent and then update phi nodes
        if len(self.items) != len(stack.items):
            raise AnalysisException("Inconsistent stack depth")
        for phi, value in zip(self.items, stack.items):
            phi.arguments.append(value)
    


class SSABlockState3(BlockState3):
    stack_size = None

    def jump_to(self, orig_cond, new_cond, bindings):
        for binding in bindings.values():
            if isinstance(binding, SSAFrameState):
                if self.stack_size is None:
                    self.stack_size = len(binding.stack)
                break
        else:
            raise AnalysisException("No stack provided")
        super(SSABlockState3, self).jump_to(orig_cond, new_cond, bindings)

    def make_phi(self, source_phi):
        proj = source_phi.sorted_proj()[0]
        if proj.ty.mode == 'M':
            return super(SSABlockState3, self).make_phi(source_phi)
        offset = self.source_block.offset
        local = []
        for index in xrange(self.graph_state.nlocals):
            local.append(self._make_phi('local', offset, [Constant(index)]))
        exc_info = [
            self._make_phi('exc_ty', offset, []),
            self._make_phi('exc_v', offset, []),
            self._make_phi('exc_tb', offset, []),
        ]
        result = self._make_phi('result', offset, [])
        mem = self._make_phi('mem', offset, [])
        stack = DumbStack([self._make_phi('stack', offset, [])
                           for _ in xrange(self.stack_size)])
        state = SSAFrameState(self, Stack(), local, exc_info, result, mem)
        self.initial_frame = state
        return state

    def merge_phi(self, target, binding):
        if target == self.initial_frame:
            return target.merge(binding)
        return super(SSABlockState3, self).merge_phi(target, binding)


class SSAOpState(OpState3):
    def __init__(self, block_state, source_op):
        super(SSAOpState, self).__init__(block_state, source_op)
        frame = block_state.get(source_op.arguments[1])
        self.stack = frame.stack
        self.frame = frame

    def yield_to_handler(self, op, value, frame):
        # yields ignore the handler stack.
        self.graph_state.graph.add_result(op)

    def no_jump(self, trans):
        # we can't remove the current block as:
        # 0: if it's the target of a jump, we need to pipe all
        # current and future jumps through to it
        #self._set_tag(None, frame, frame.mem, self._get_true(),
        #              **frame_attrs)

        frame = frame or self.frame
        C, V, M = self._get_proj_family(tag)
        def pre_trans(bindings):
            matches = []
            for target in self.block_state.phi_nodes:
                for source in target.arguments:
                    if source in bindings:
                        matches.append(bindings[source])
                        break
                else:
                    raise AnalysisException(
                        "Could not find a binding for %r in %r" %
                        (target, bindings)
                    )
            
            frame_attrs = trans(*matches)
            return {
                M: [self.block_state.get(mem) for mem in self.source_op.mem],
                #V: matches[],
            }
        self.graph_state.skip_block(self.block_state.source_block, C,
                                    pre_trans)

    def no_jump(self, trans):
        sop = self.source_op
        mem = [self.block_state.get(mem) for mem in sop.mem]
        #arguments = [self.block_state.get(arg) for arg in sop.arguments]
        self.set_tag(None, self.op(sop.opname, mem,
                                   [self.block_state.get(sop.arguments[1])]))

    def set_tag(self, tag, op, **frame_attrs):
        frame = self.block_state.get(self.source_op.arguments[1])
        if tag is None:
            M, C = op.proj['M'], op.proj['C']
        else:
            M, C = op.proj[tag + '_M'], op.proj[tag + '_C']
        self._set_tag(frame, M, C, **frame_attrs)

    def custom_set(self, tag, mem, edge, **frame_attrs):
        self._set_tag(tag, self.frame, mem, edge, **frame_attrs)

    def _set_tag(self, tag, frame, mem, edge, **frame_attrs):
        C, V, M = self._get_proj_family(tag)
        self.parent.add_jump(C, edge, {
            V: frame.set_to(**frame_attrs),
            M: mem,
        })

    def set_next(self, op, tag='C', **frame_attrs):
        self._set_tag(None, self.frame, op.proj_M, op.proj[tag], **frame_attrs)

    def set_alt(self, frame, mem, edge, stack=None, **frame_attrs):
        self._set_tag('alt', frame, mem, edge, stack=stack, **frame_attrs)

    def _get_proj_family(self, family=None):
        for mode in 'CVM':
            yield self.source_op['%s_%s' % (family, mode) if family else mode]

    def _get_true(self):
        true = self.op('entangled.true', [], []).proj_C

    # def jump_to_handler(self, frame, mem, why, exc_info=None, **frame_attrs):
    #     # the source operation already has connections to the handlers
    #     # - we need to look them up and bind them in the function's
    #     # context
    #     tag = why.projname
    #     self._set_tag(frame, op.proj[tag + '_M'], op.proj[tag + '_C'],
    #                   tag=tag, **frame_attrs)

    def unroll(self, op, tag='exc', unroller='SApplicationException'):
        unroller_v = self.op(unroller, [op.proj_M], [op.proj[tag + '_EX']])
        self._set_tag(tag, self.frame,
                      [unroller_v.proj_M], [op.proj[tag + '_C']],
                      unroller=unroller_v)

    def continue_(self, op):
        self.unroll(op, 'alt', 'SContinueLoop')

    def break_(self, op):
        self.unroll(op, 'break', 'SBreakLoop')

    def return_(self, op):
        self.unroll(op, 'return', 'SReturnValue')

    # def may_raise(self, frame, op, mem=None):
    #     if mem is None:
    #         mem = op.proj_EX_M
    #     exc_info = [op.proj_EX_TY, op.proj_EX_V, op.proj_EX_TB]
    #     self.jump_to_handler(frame, mem, self.WHY_RAISE, exc_info)
    #     unroller = self.op('SApplicationException', [mem], [op])
    #     self._set_tag('exit', self.frame, mem, op.proj_EX_C,
    #                   unroller=unroller):

    def add_scope(self, orig, value):
        self.graph_state.scope[orig] = value


class SSAFrameState(FrameState):
    def merge(self, other_frame):
        self.stack.merge(other_frame.stack)
        for phi, source in zip(self.local, other_frame.local):
            phi.arguments.append(source)
        for phi, source in zip(self.exc_info, other_frame.exc_info):
            phi.arguments.append(source)
        self.result.arguments.append(other_frame.result)
        self.mem.arguments.append(other_frame.mem)


class SSAAnalyser(BaseAnalyser):
    def make_graph_state(self, source_graph):
        """Create an IGraphState for analysis.
        """
        return GraphState(self, source_graph)

    def make_block_state(self, graph_state, source_block):
        """Create an IBlockState for later analysis.
        """
        return SSABlockState(graph_state, source_block)

    def make_op_state(self, block_state, source_op):
        """Create an IOpState state for analysis.
        """
        return OpState(block_state, source_op)

    
